import os
import posixpath
import numpy as np
from silx.io import get_data
from ...utils import copy_dict_items, compare_dicts
from ...io.utils import hdf5_entry_exists, get_h5_value
from ...io.reader import import_h5_to_dict
from ...resources.utils import extract_parameters, get_values_from_file
from ...resources.nxflatfield import update_dataset_info_flats_darks
from ..estimators import COREstimator, SinoCOREstimator, CompositeCOREstimator
from ..processconfig import ProcessConfigBase
from .nabu_config import nabu_config, renamed_keys
from .dataset_validator import FullFieldDatasetValidator


class ProcessConfig(ProcessConfigBase):
    default_nabu_config = nabu_config
    config_renamed_keys = renamed_keys

    def _update_dataset_info_with_user_config(self):
        """
        Update the 'dataset_info' (DatasetAnalyzer class instance) data structure with options from user configuration.
        """
        self.logger.debug("Updating dataset information with user configuration")
        if self.dataset_info.kind == "hdf5":
            update_dataset_info_flats_darks(
                self.dataset_info,
                processes_file=self.nabu_config["preproc"]["processes_file"],
                output_dir=self.nabu_config["output"]["location"],
                flatfield_mode=self.nabu_config["preproc"]["flatfield"],
            )
        self._get_rotation_axis_position()
        self._update_rotation_angles()
        self._get_translation_file("reconstruction", "translation_movements_file", "translations")
        self._get_translation_file("phase", "ctf_translations_file", "ctf_translations")


    def _get_translation_file(self, config_section, config_key, dataset_info_attr):
        transl_file = self.nabu_config[config_section][config_key]
        if transl_file in (None, ''):
            return
        translations = None
        if transl_file is not None and "://" not in transl_file:
            try:
                translations = get_values_from_file(
                    transl_file, shape=(self.dataset_info.n_angles, 2)
                ).astype(np.float32)
            except ValueError:
                print("Something wrong with translation_movements_file %s" % transl_file)
                raise
        else:
            try:
                translations = get_data(transl_file)
            except:
                print("Something wrong with translation_movements_file %s" % transl_file)
                raise
        setattr(self.dataset_info, dataset_info_attr, translations)

    def _get_rotation_axis_position(self):
        super()._get_rotation_axis_position()
        rec_params = self.nabu_config["reconstruction"]
        axis_correction_file = rec_params["axis_correction_file"]
        axis_correction = None
        if axis_correction_file is not None:
            try:
                axis_correction = get_values_from_file(
                    axis_correction_file, n_values=self.dataset_info.n_angles
                ).astype(np.float32)
            except ValueError:
                print("Something wrong with axis correction file %s" % axis_correction_file)
                raise
        self.dataset_info.axis_correction = axis_correction


    def _update_rotation_angles(self):
        rec_params = self.nabu_config["reconstruction"]
        n_angles = self.dataset_info.n_angles
        angles_file = rec_params["angles_file"]
        if angles_file is not None:
            try:
                angles = get_values_from_file(angles_file, n_values=n_angles)
                angles = np.deg2rad(angles)
            except ValueError:
                self.logger.fatal("Something wrong with angle file %s" % angles_file)
                raise
            self.dataset_info.rotation_angles = angles
        elif self.dataset_info.rotation_angles is None:
            angles_range_txt = "[0, 180[ degrees"
            if rec_params["enable_halftomo"]:
                angles_range_txt = "[0, 360] degrees"
                angles = np.linspace(0, 2*np.pi, n_angles, True)
            else:
                angles = np.linspace(0, np.pi, n_angles, False)
            self.logger.warning(
                "No information was found on rotation angles. Using default %s (half tomo is %s)"
                % (angles_range_txt, {0: "OFF", 1: "ON"}[int(self.do_halftomo)])
            )
            self.dataset_info.rotation_angles = angles
        angles = self.dataset_info.rotation_angles
        angles += np.deg2rad(rec_params["angle_offset"])
        self.dataset_info.reconstruction_angles = angles


    def _get_cor(self):
        cor = self.nabu_config["reconstruction"]["rotation_axis_position"]
        if isinstance(cor, str): # auto-CoR
            cor_slice = self.nabu_config["reconstruction"]["cor_slice"]
            if cor_slice is not None or cor == "sino-coarse-to-fine":
                subsampling = extract_parameters(
                    self.nabu_config["reconstruction"]["cor_options"]
                ).get("subsampling", 10)
                self.corfinder = SinoCOREstimator(
                    self.dataset_info,
                    cor_slice or 0,
                    subsampling=subsampling,
                    do_flatfield=bool(self.nabu_config["preproc"]["flatfield"]),
                    cor_options=self.nabu_config["reconstruction"]["cor_options"],
                    logger=self.logger
                )
            elif cor == "composite-coarse-to-fine":
                self.corfinder = CompositeCOREstimator(
                    self.dataset_info,
                    cor_options=self.nabu_config["reconstruction"]["cor_options"],
                    logger=self.logger
                )
            else:
                self.corfinder = COREstimator(
                    self.dataset_info,
                    halftomo=self.nabu_config["reconstruction"]["enable_halftomo"],
                    do_flatfield=bool(self.nabu_config["preproc"]["flatfield"]),
                    cor_options=self.nabu_config["reconstruction"]["cor_options"],
                    logger=self.logger
                )
            cor = self.corfinder.find_cor(method=cor)
            self.logger.info("Estimated center of rotation: %.3f" % cor)
        self.dataset_info.axis_position = cor


    @property
    def do_halftomo(self):
        """
        Return True if the current dataset is to be reconstructed using 'half-acquisition' setting.
        """
        enable_halftomo = self.nabu_config["reconstruction"]["enable_halftomo"]
        is_halftomo_dataset = self.dataset_info.is_halftomo
        if enable_halftomo == "auto":
            if is_halftomo_dataset is None:
                raise ValueError("enable_halftomo was set to 'auto', but information on field of view was not found. Please set either 0 or 1 for enable_halftomo")
            return is_halftomo_dataset
        return enable_halftomo


    def _coupled_validation(self):
        self.logger.debug("Doing coupled validation")
        self._dataset_validator = FullFieldDatasetValidator(self.nabu_config, self.dataset_info)


    # TODO update behavior and remove this function once GroupedPipeline cuda backend is implemented
    def get_radios_rotation_mode(self):
        """
        Determine whether projections are to be rotated, and if so, when they are to be rotated.

        Returns
        -------
        method: str or None
            Rotation method: one of the values of `nabu.resources.params.radios_rotation_mode`
        """
        user_rotate_projections = self.nabu_config["preproc"]["rotate_projections"]
        tilt = self.dataset_info.detector_tilt
        phase_method = self.nabu_config["phase"]["method"]
        do_ctf = phase_method == "CTF"
        do_pag = phase_method == "paganin"
        do_unsharp = self.nabu_config["phase"]["unsharp_method"] is not None and self.nabu_config["phase"]["unsharp_coeff"] > 0
        if user_rotate_projections is None and tilt is None:
            return None
        if do_ctf:
            return "full"
        # TODO "chunked" rotation is done only when using a "processing margin"
        # For now the processing margin is enabled only if phase or unsharp is enabled.
        # We can either
        #   - Enable processing margin if rotating projections is needed (more complicated to implement)
        #   - Always do "full" rotation (simpler to implement, at the expense of performances)
        if do_pag or do_unsharp:
            return "chunk"
        else:
            return "full"


    def _build_processing_steps(self):
        nabu_config = self.nabu_config
        dataset_info = self.dataset_info
        binning = (nabu_config["dataset"]["binning"], nabu_config["dataset"]["binning_z"])
        tasks = []
        options = {}

        #
        # Dataset / Get data
        #
        # First thing to do is to get the data (radios or sinograms)
        # For now data is assumed to be on disk (see issue #66).
        tasks.append("read_chunk")
        options["read_chunk"] = {
            "files": dataset_info.projections,
            "sub_region": None,
            "binning": binning,
            "dataset_subsampling": nabu_config["dataset"]["projections_subsampling"]
        }
        #
        # Flat-field
        #
        if nabu_config["preproc"]["flatfield"]:
            tasks.append("flatfield")
            options["flatfield"] = {
                #  ChunkReader handles binning/subsampling by itself,
                # but FlatField needs "real" indices (after binning/subsampling)
                "projs_indices": dataset_info._projs_indices_subsampled,
                "binning": binning,
                "do_flat_distortion": nabu_config["preproc"]["flat_distortion_correction_enabled"],
                "flat_distortion_params": extract_parameters(nabu_config["preproc"]["flat_distortion_params"]),
            }
        #
        # Spikes filter
        #
        if nabu_config["preproc"]["ccd_filter_enabled"]:
            tasks.append("ccd_correction")
            options["ccd_correction"] = {
                "type": "median_clip", # only one available for now
                "median_clip_thresh": nabu_config["preproc"]["ccd_filter_threshold"],
            }
        #
        # Double flat field
        #
        if nabu_config["preproc"]["double_flatfield_enabled"]:
            tasks.append("double_flatfield")
            options["double_flatfield"] = {
                "sigma": nabu_config["preproc"]["dff_sigma"],
                "processes_file": nabu_config["preproc"]["processes_file"],
            }
        #
        # Radios rotation (do it here if possible)
        #
        if self.get_radios_rotation_mode() == "chunk":
            tasks.append("rotate_projections")
            options["rotate_projections"] = {
                "angle": nabu_config["preproc"]["rotate_projections"] or dataset_info.detector_tilt,
                "center": nabu_config["preproc"]["rotate_projections_center"],
                "mode": "chunk",
            }
        #
        #
        # Phase retrieval
        #
        if nabu_config["phase"]["method"] is not None:
            tasks.append("phase")
            options["phase"] = copy_dict_items(
                nabu_config["phase"], ["method", "delta_beta", "margin", "padding_type"]
            )
            options["phase"].update({
                "energy_kev": dataset_info.energy,
                "distance_cm": dataset_info.distance * 1e2,
                "distance_m": dataset_info.distance,
                "pixel_size_microns": dataset_info.pixel_size,
                "pixel_size_m": dataset_info.pixel_size * 1e-6,
            })
            if binning != (1, 1):
                options["phase"]["delta_beta"] /= (binning[0] * binning[1])
            if options["phase"]["method"] == "CTF":
                self._get_ctf_parameters(options["phase"])
        #
        # Unsharp
        #
        if nabu_config["phase"]["unsharp_method"] is not None and nabu_config["phase"]["unsharp_coeff"] > 0:
            tasks.append("unsharp_mask")
            options["unsharp_mask"] = copy_dict_items(
                nabu_config["phase"], ["unsharp_coeff", "unsharp_sigma", "unsharp_method"]
            )
        #
        # -logarithm
        #
        if nabu_config["preproc"]["take_logarithm"]:
            tasks.append("take_log")
            options["take_log"] = copy_dict_items(nabu_config["preproc"], ["log_min_clip", "log_max_clip"])
        #
        # Radios rotation (do it here if mode=="full")
        #
        if self.get_radios_rotation_mode() == "full":
            tasks.append("rotate_projections")
            options["rotate_projections"] = {
                "angle": nabu_config["preproc"]["rotate_projections"] or dataset_info.detector_tilt,
                "center": nabu_config["preproc"]["rotate_projections_center"],
                "mode": "full",
            }
        #
        # Translation movements
        #
        translations = dataset_info.translations
        if translations is not None:
            tasks.append("radios_movements")
            options["radios_movements"] = {
                "translation_movements": dataset_info.translations
            }
        #
        # Sinogram normalization (before half-tomo)
        #
        if nabu_config["preproc"]["sino_normalization"] is not None:
            tasks.append("sino_normalization")
            options["sino_normalization"] = {
                "method": nabu_config["preproc"]["sino_normalization"]
            }
        #
        # Sinogram-based rings artefacts removal
        #
        if nabu_config["preproc"]["sino_rings_correction"]:
            tasks.append("sino_rings_correction")
            options["sino_rings_correction"] = {
                "user_options": nabu_config["preproc"]["sino_rings_options"],
            }
        #
        # Reconstruction
        #
        if nabu_config["reconstruction"]["method"] is not None:
            tasks.append("build_sino")
            options["build_sino"] = copy_dict_items(
                nabu_config["reconstruction"],
                ["rotation_axis_position", "start_x", "end_x", "start_y", "end_y", "start_z", "end_z"]
            )
            options["build_sino"]["axis_correction"] = dataset_info.axis_correction
            tasks.append("reconstruction")
            # Iterative is not supported through configuration file for now.
            options["reconstruction"] = copy_dict_items(
                nabu_config["reconstruction"],
                [
                    "method", "rotation_axis_position", "fbp_filter_type",
                    "padding_type", "start_x", "end_x", "start_y", "end_y", "start_z", "end_z",
                    "centered_axis", "clip_outer_circle"
                ]
            )
            rec_options = options["reconstruction"]
            rec_options["rotation_axis_position"] = dataset_info.axis_position
            rec_options["enable_halftomo"] = self.do_halftomo
            options["build_sino"]["rotation_axis_position"] = dataset_info.axis_position
            options["build_sino"]["enable_halftomo"] = self.do_halftomo
            rec_options["axis_correction"] = dataset_info.axis_correction
            rec_options["angles"] = dataset_info.reconstruction_angles
            rec_options["radio_dims_y_x"] = dataset_info.radio_dims[::-1]
            rec_options["pixel_size_cm"] = dataset_info.pixel_size * 1e-4 # pix size is in microns
            if self.do_halftomo:
                rec_options["angles"] = rec_options["angles"][:(rec_options["angles"].size + 1)//2]
                cor_i = int(round(rec_options["rotation_axis_position"]))
                # New keys
                rec_options["rotation_axis_position_halftomo"] = (2*cor_i-1)/2.
            # New key
            rec_options["cor_estimated_auto"] = isinstance(nabu_config["reconstruction"]["rotation_axis_position"], str)
        #
        # Histogram
        #
        if nabu_config["postproc"]["output_histogram"]:
            tasks.append("histogram")
            options["histogram"] = copy_dict_items(
                nabu_config["postproc"], ["histogram_bins"]
            )
        #
        # Save
        #
        if nabu_config["output"]["location"] is not None:
            tasks.append("save")
            options["save"] = copy_dict_items(
                nabu_config["output"], list(nabu_config["output"].keys())
            )
            options["save"]["overwrite"] = nabu_config["output"]["overwrite_results"]

        self.processing_steps = tasks
        self.processing_options = options
        if set(self.processing_steps) != set(self.processing_options.keys()):
            raise ValueError("Something wrong with process_config: options do not correspond to steps")
        # Add check
        if set(self.processing_steps) != set(self.processing_options.keys()):
            raise ValueError("Something wrong when building processing steps")
        #
        self._configure_save_steps()
        self._configure_resume()


    def _get_ctf_parameters(self, phase_options):
        dataset_info = self.dataset_info
        user_phase_options = self.nabu_config["phase"]

        ctf_geom = extract_parameters(user_phase_options["ctf_geometry"])
        ctf_advanced_params = extract_parameters(user_phase_options["ctf_advanced_params"])

        # z1_vh
        z1_v = ctf_geom["z1_v"]
        z1_h = ctf_geom["z1_h"]
        z1_vh = None
        if z1_h is None and z1_v is None:
            # parallel beam
            z1_vh = None
        elif (z1_v is None) ^ (z1_h is None):
            # only one is provided: source-sample distance
            z1_vh = z1_v or z1_h
        if z1_h is not None and z1_v is not None:
            # distance of the vertically focused source (horizontal line) and the horizontaly focused source (vertical line)
            # for KB mirrors
            z1_vh = (z1_v, z1_h)
        # pix_size_det
        pix_size_det = ctf_geom["detec_pixel_size"] or dataset_info.pixel_size * 1e-6
        # wavelength
        wavelength = 1.23984199e-9 / dataset_info.energy

        phase_options["ctf_geo_pars"] = {
            "z1_vh": z1_vh,
            "z2": phase_options["distance_m"],
            "pix_size_det": pix_size_det,
            "wavelength": wavelength,
            "magnification": bool(ctf_geom["magnification"]),
            "length_scale": ctf_advanced_params["length_scale"]
        }
        phase_options["ctf_lim1"] = ctf_advanced_params["lim1"]
        phase_options["ctf_lim2"] = ctf_advanced_params["lim2"]
        phase_options["ctf_normalize_by_mean"] = ctf_advanced_params["normalize_by_mean"]


    def _configure_save_steps(self):
        self._dump_sinogram = False
        steps_to_save = self.nabu_config["pipeline"]["save_steps"]
        if steps_to_save in (None, ""):
            self.steps_to_save = []
            return
        steps_to_save = [s.strip() for s in steps_to_save.split(",")]
        for step in self.processing_steps:
            step = step.strip()
            if step in steps_to_save:
                self.processing_options[step]["save"] = True
                self.processing_options[step]["save_steps_file"] = self.get_save_steps_file(step_name=step)
        # "sinogram" is a special keyword, not explicitly in the processing steps
        if "sinogram" in steps_to_save:
            self._dump_sinogram = True
            self._dump_sinogram_file = self.get_save_steps_file(step_name="sinogram")
        self.steps_to_save = steps_to_save


    def _get_dump_file_and_h5_path(self):
        resume_from = self.resume_from_step
        process_file = self.get_save_steps_file(step_name=resume_from)
        if not os.path.isfile(process_file):
            self.logger.error(
                "Cannot resume processing from step '%s': no such file %s" % (resume_from, process_file)
            )
            return None, None
        h5_entry = self.dataset_info.hdf5_entry or "entry"
        process_h5_path = posixpath.join(
            h5_entry,
            resume_from,
            "results/data"
        )
        if not hdf5_entry_exists(process_file, process_h5_path):
            self.logger.error(
                "Could not find data in %s in file %s" % (process_h5_path, process_file)
            )
            process_h5_path = None
        return process_file, process_h5_path


    def _configure_resume(self):
        resume_from = self.nabu_config["pipeline"]["resume_from_step"]
        if resume_from in (None, ""):
            self.resume_from_step = None
            return
        resume_from = resume_from.strip(" ,;")
        self.resume_from_step = resume_from

        processing_steps = self.processing_steps
        # special case: resume from sinogram
        if resume_from == "sinogram":
            if "build_sino" not in processing_steps:
                msg = "Cannot resume processing from step 'sinogram': reconstruction is disabled with this configuration"
                self.logger.fatal(msg)
                raise ValueError(msg)
            idx = processing_steps.index("build_sino") # disable up to 'build_sino', not included
        #
        elif resume_from in processing_steps:
            idx = processing_steps.index(resume_from) + 1 # disable up to resume_from, included
        else:
            msg = "Cannot resume processing from step '%s': no such step in the current configuration" % resume_from
            self.logger.error(msg)
            self.resume_from_step = None
            return

        # Get corresponding file and h5 path
        process_file, process_h5_path = self._get_dump_file_and_h5_path()
        if process_file is None or process_h5_path is None:
            self.resume_from_step = None
            return
        dump_info = self._check_dump_file(process_file, raise_on_error=False)
        if dump_info is None:
            self.logger.error(
                "Cannot resume from step %s: cannot use file %s" % (resume_from, process_file)
            )
            self.resume_from_step = None
            return
        dump_start_z, dump_end_z = dump_info

        # Disable steps
        steps_to_disable = processing_steps[1:idx]
        self.logger.debug("Disabling steps %s" % str(steps_to_disable))
        for step_name in steps_to_disable:
            processing_steps.remove(step_name)
            self.processing_options.pop(step_name)

        # Update configuration
        self.logger.info(
            "Processing will be resumed from step '%s' using file %s"
            % (resume_from, process_file)
        )
        self._old_read_chunk = self.processing_options["read_chunk"]
        self.processing_options["read_chunk"] = {
            "process_file": process_file,
            "process_h5_path": process_h5_path,
            "step_name": resume_from,
            "dump_start_z": dump_start_z,
            "dump_end_z": dump_end_z
        }
        # Dont dump a step if we resume from this step
        if resume_from in self.steps_to_save:
            self.logger.warning(
                "Processing is resumed from step '%s'. This step won't be dumped to a file" % resume_from
            )
            self.steps_to_save.remove(resume_from)
            if resume_from == "sinogram":
                self._dump_sinogram = False
            else:
                self.processing_options[resume_from].pop("save")


    def _check_dump_file(self, process_file, raise_on_error=False):
        """
        Return (start_z, end_z) on success
        Return None on failure
        """
        # Ensure data in the file correspond to what is currently asked
        if self.resume_from_step is None:
            return None

        # Check dataset shape/start_z/end_z
        rec_cfg_h5_path = posixpath.join(
            self.dataset_info.hdf5_entry or "entry",
            self.resume_from_step,
            "configuration/nabu_config/reconstruction"
        )
        dump_start_z = get_h5_value(process_file, posixpath.join(rec_cfg_h5_path, "start_z"))
        dump_end_z = get_h5_value(process_file, posixpath.join(rec_cfg_h5_path, "end_z"))
        start_z, end_z = self.nabu_config["reconstruction"]["start_z"], self.nabu_config["reconstruction"]["end_z"]
        if not (dump_start_z <= start_z and end_z <= dump_end_z):
            msg = "File %s was built with start_z=%d, end_z=%d but current configuration asks for start_z=%d, end_z=%d" % (process_file, dump_start_z, dump_end_z, start_z, end_z)
            if not raise_on_error:
                self.logger.error(msg)
                return None
            self.logger.fatal(msg)
            raise ValueError(msg)

        # Check parameters other than reconstruction
        filedump_nabu_config = import_h5_to_dict(
            process_file,
            posixpath.join(
                self.dataset_info.hdf5_entry or "entry",
                self.resume_from_step,
                "configuration/nabu_config"
            )
        )
        sections_to_ignore = ["reconstruction", "output", "pipeline"]
        for section in sections_to_ignore:
            filedump_nabu_config[section] = self.nabu_config[section]
        # special case of the "save_steps process"
        # filedump_nabu_config["pipeline"]["save_steps"] = self.nabu_config["pipeline"]["save_steps"]

        diff = compare_dicts(filedump_nabu_config, self.nabu_config)
        if diff is not None:
            msg = "Nabu configuration in file %s differ from the current one: %s" % (process_file, diff)
            if not raise_on_error:
                self.logger.error(msg)
                return None
            self.logger.fatal(msg)
            raise ValueError(msg)
        #

        return (dump_start_z, dump_end_z)


    def get_save_steps_file(self, step_name=None):
        if self.nabu_config["pipeline"]["steps_file"] not in (None, ""):
            return self.nabu_config["pipeline"]["steps_file"]
        nabu_save_options = self.nabu_config["output"]
        output_dir = nabu_save_options["location"]
        file_prefix = step_name + "_" + nabu_save_options["file_prefix"]
        fname = os.path.join(output_dir, file_prefix) + ".hdf5"
        return fname



# import math
import logging
import numpy as np
from numpy.polynomial.polynomial import Polynomial
from scipy.ndimage.filters import median_filter
import scipy.fft
from ..utils import previouspow2
from ..misc import fourier_filters
from ..resources.logger import LoggerOrPrint
try:
    import matplotlib.pyplot as plt
    __have_matplotlib__ = True
except ImportError:
    logging.getLogger(__name__).warning("Matplotlib not available. Plotting disabled")
    plt = None
    __have_matplotlib__ = False

try:
    from tqdm import tqdm
    def progress_bar(x, verbose=True):
        if verbose:
            return tqdm(x)
        else:
            return x
except ImportError:
    def progress_bar(x, verbose=True):
        return x

local_fftn = scipy.fft.rfftn
local_ifftn = scipy.fft.irfftn

class AlignmentBase:
    def __init__(
        self, vert_fft_width=False, horz_fft_width=False, verbose=False, logger=None, data_type=np.float32,
    ):
        """
        Alignment basic functions.

        Parameters
        ----------
        vert_fft_width: boolean, optional
            If True, restrict the vertical size to a power of 2:

            >>> new_v_dim = 2 ** math.floor(math.log2(v_dim))

        horz_fft_width: boolean, optional
            If True, restrict the horizontal size to a power of 2:

            >>> new_h_dim = 2 ** math.floor(math.log2(h_dim))

        verbose: boolean, optional
            When True it will produce verbose output, including plots.
        data_type: `numpy.float32`
            Computation data type.
        """

        self._init_parameters(vert_fft_width, horz_fft_width, verbose, logger, data_type)
        self._plot_windows = {}

    def _init_parameters(self, vert_fft_width, horz_fft_width, verbose, logger, data_type):
        self.logger = LoggerOrPrint(logger)
        self.truncate_vert_pow2 = vert_fft_width
        self.truncate_horz_pow2 = horz_fft_width

        if verbose and not __have_matplotlib__:
            self.logger.warning("Matplotlib not available. Plotting disabled, despite being activated by user")
            verbose = False
        self.verbose = verbose
        self.data_type = data_type

    @staticmethod
    def _check_img_stack_size(img_stack: np.ndarray, img_pos: np.ndarray):
        shape_stack = np.squeeze(img_stack).shape
        shape_pos = np.squeeze(img_pos).shape
        if not len(shape_stack) == 3:
            raise ValueError(
                "A stack of 2-dimensional images is required. Shape of stack: %s" % (" ".join(("%d" % x for x in shape_stack)))
            )
        if not len(shape_pos) == 1:
            raise ValueError(
                "Positions need to be a 1-dimensional array. Shape of the positions variable: %s"
                % (" ".join(("%d" % x for x in shape_pos)))
            )
        if not shape_stack[0] == shape_pos[0]:
            raise ValueError(
                "The same number of images and positions is required."
                + " Shape of stack: %s, shape of positions variable: %s"
                % (" ".join(("%d" % x for x in shape_stack)), " ".join(("%d" % x for x in shape_pos)),)
            )

    @staticmethod
    def _check_img_pair_sizes(img_1: np.ndarray, img_2: np.ndarray):
        shape_1 = np.squeeze(img_1).shape
        shape_2 = np.squeeze(img_2).shape
        if not len(shape_1) == 2:
            raise ValueError("Images need to be 2-dimensional. Shape of image #1: %s" % (" ".join(("%d" % x for x in shape_1))))
        if not len(shape_2) == 2:
            raise ValueError("Images need to be 2-dimensional. Shape of image #2: %s" % (" ".join(("%d" % x for x in shape_2))))
        if not np.all(shape_1 == shape_2):
            raise ValueError(
                "Images need to be of the same shape. Shape of image #1: %s, image #2: %s"
                % (" ".join(("%d" % x for x in shape_1)), " ".join(("%d" % x for x in shape_2)),)
            )

    @staticmethod
    def refine_max_position_2d(f_vals: np.ndarray, fy=None, fx=None):
        """Computes the sub-pixel max position of the given function sampling.

        Parameters
        ----------
        f_vals: numpy.ndarray
            Function values of the sampled points
        fy: numpy.ndarray, optional
            Vertical coordinates of the sampled points
        fx: numpy.ndarray, optional
            Horizontal coordinates of the sampled points

        Raises
        ------
        ValueError
            In case position and values do not have the same size, or in case
            the fitted maximum is outside the fitting region.

        Returns
        -------
        tuple(float, float)
            Estimated (vertical, horizontal) function max, according to the
            coordinates in fy and fx.
        """
        if not (len(f_vals.shape) == 2):
            raise ValueError(
                "The fitted values should form a 2-dimensional array. Array of shape: [%s] was given."
                % (" ".join(("%d" % s for s in f_vals.shape)))
            )
        if fy is None:
            fy_half_size = (f_vals.shape[0] - 1) / 2
            fy = np.linspace(-fy_half_size, fy_half_size, f_vals.shape[0])
        elif not (len(fy.shape) == 1 and np.all(fy.size == f_vals.shape[0])):
            raise ValueError(
                "Vertical coordinates should have the same length as values matrix. Sizes of fy: %d, f_vals: [%s]"
                % (fy.size, " ".join(("%d" % s for s in f_vals.shape)))
            )
        if fx is None:
            fx_half_size = (f_vals.shape[1] - 1) / 2
            fx = np.linspace(-fx_half_size, fx_half_size, f_vals.shape[1])
        elif not (len(fx.shape) == 1 and np.all(fx.size == f_vals.shape[1])):
            raise ValueError(
                "Horizontal coordinates should have the same length as values matrix. Sizes of fx: %d, f_vals: [%s]"
                % (fx.size, " ".join(("%d" % s for s in f_vals.shape)))
            )

        fy, fx = np.meshgrid(fy, fx, indexing="ij")
        fy = fy.flatten()
        fx = fx.flatten()
        coords = np.array([np.ones(f_vals.size), fy, fx, fy * fx, fy ** 2, fx ** 2])

        coeffs = np.linalg.lstsq(coords.T, f_vals.flatten(), rcond=None)[0]

        # For a 1D parabola `f(x) = ax^2 + bx + c`, the vertex position is:
        # x_v = -b / 2a. For a 2D parabola, the vertex position is:
        # (y, x)_v = - b / A, where:
        A = [[2 * coeffs[4], coeffs[3]], [coeffs[3], 2 * coeffs[5]]]
        b = coeffs[1:3]
        vertex_yx = np.linalg.lstsq(A, -b, rcond=None)[0]

        vertex_min_yx = [np.min(fy), np.min(fx)]
        vertex_max_yx = [np.max(fy), np.max(fx)]
        if np.any(vertex_yx < vertex_min_yx) or np.any(vertex_yx > vertex_max_yx):
            raise ValueError(
                "Fitted (y: {}, x: {}) positions are outside the input margins y: [{}, {}], and x: [{}, {}]".format(
                    vertex_yx[0], vertex_yx[1], vertex_min_yx[0], vertex_max_yx[0], vertex_min_yx[1], vertex_max_yx[1],
                )
            )
        return vertex_yx

    @staticmethod
    def refine_max_position_1d(f_vals, fx=None, return_vertex_val=False):
        """Computes the sub-pixel max position of the given function sampling.

        Parameters
        ----------
        f_vals: numpy.ndarray
            Function values of the sampled points
        fx: numpy.ndarray, optional
            Coordinates of the sampled points
        return_vertex_val: boolean, option
            Enables returning the vertex values. Defaults to False.

        Raises
        ------
        ValueError
            In case position and values do not have the same size, or in case
            the fitted maximum is outside the fitting region.

        Returns
        -------
        float
            Estimated function max, according to the coordinates in fx.
        """
        if not len(f_vals.shape) in (1, 2):
            raise ValueError(
                "The fitted values should be either one or a collection of 1-dimensional arrays. Array of shape: [%s] was given."
                % (" ".join(("%d" % s for s in f_vals.shape)))
            )
        num_vals = f_vals.shape[0]

        if fx is None:
            fx_half_size = (num_vals - 1) / 2
            fx = np.linspace(-fx_half_size, fx_half_size, num_vals)
        else:
            fx = np.squeeze(fx)
            if not (len(fx.shape) == 1 and np.all(fx.size == num_vals)):
                raise ValueError(
                    "Base coordinates should have the same length as values array. Sizes of fx: %d, f_vals: %d"
                    % (fx.size, num_vals)
                )

        if len(f_vals.shape) == 1:
            # using Polynomial.fit, because supposed to be more numerically
            # stable than previous solutions (according to numpy).
            poly = Polynomial.fit(fx, f_vals, deg=2)
            coeffs = poly.convert().coef
        else:
            coords = np.array([np.ones(num_vals), fx, fx ** 2])
            coeffs = np.linalg.lstsq(coords.T, f_vals, rcond=None)[0]

        # For a 1D parabola `f(x) = c + bx + ax^2`, the vertex position is:
        # x_v = -b / 2a.
        vertex_x = -coeffs[1, :] / (2 * coeffs[2, :])

        vertex_min_x = np.min(fx)
        vertex_max_x = np.max(fx)
        lower_bound_ok = vertex_min_x < vertex_x
        upper_bound_ok = vertex_x < vertex_max_x
        if not np.all(lower_bound_ok * upper_bound_ok):
            if len(f_vals.shape) == 1:
                message = "Fitted position {} is outide the input margins [{}, {}]".format(vertex_x, vertex_min_x, vertex_max_x)
            else:
                message = "Fitted positions outide the input margins [{}, {}]: {} below and {} above".format(
                    vertex_min_x, vertex_max_x, np.sum(1 - lower_bound_ok), np.sum(1 - upper_bound_ok),
                )
            raise ValueError(message)
        if return_vertex_val:
            vertex_val = coeffs[0, :] + vertex_x * coeffs[1, :] / 2
            return vertex_x, vertex_val
        else:
            return vertex_x

    @staticmethod
    def extract_peak_region_2d(cc, peak_radius=1, cc_vs=None, cc_hs=None):
        """
        Extracts a region around the maximum value.

        Parameters
        ----------
        cc: numpy.ndarray
            Correlation image.
        peak_radius: int, optional
            The l_inf radius of the area to extract around the peak. The default is 1.
        cc_vs: numpy.ndarray, optional
            The vertical coordinates of `cc`. The default is None.
        cc_hs: numpy.ndarray, optional
            The horizontal coordinates of `cc`. The default is None.

        Returns
        -------
        f_vals: numpy.ndarray
            The extracted function values.
        fv: numpy.ndarray
            The vertical coordinates of the extracted values.
        fh: numpy.ndarray
            The horizontal coordinates of the extracted values.
        """
        img_shape = np.array(cc.shape)
        # get pixel having the maximum value of the correlation array
        pix_max_corr = np.argmax(cc)
        pv, ph = np.unravel_index(pix_max_corr, img_shape)

        # select a n x n neighborhood for the sub-pixel fitting (with wrapping)
        pv = np.arange(pv - peak_radius, pv + peak_radius + 1) % img_shape[-2]
        ph = np.arange(ph - peak_radius, ph + peak_radius + 1) % img_shape[-1]

        # extract the (v, h) pixel coordinates
        fv = None if cc_vs is None else cc_vs[pv]
        fh = None if cc_hs is None else cc_hs[ph]

        # extract the correlation values
        pv, ph = np.meshgrid(pv, ph, indexing="ij")
        f_vals = cc[pv, ph]

        return (f_vals, fv, fh)

    @staticmethod
    def extract_peak_regions_1d(cc, axis=-1, peak_radius=1, cc_coords=None):
        """
        Extracts a region around the maximum value.

        Parameters
        ----------
        cc: numpy.ndarray
            Correlation image.
        axis: int, optional
            Find the max values along the specified direction. The default is -1.
        peak_radius: int, optional
            The l_inf radius of the area to extract around the peak. The default is 1.
        cc_coords: numpy.ndarray, optional
            The coordinates of `cc` along the selected axis. The default is None.

        Returns
        -------
        f_vals: numpy.ndarray
            The extracted function values.
        fc_ax: numpy.ndarray
            The coordinates of the extracted values, along the selected axis.
        """
        if len(cc.shape) == 1:
            cc = cc[None, ...]
        img_shape = np.array(cc.shape)
        if not (len(img_shape) == 2):
            raise ValueError(
                "The input image should be either a 1 or 2-dimensional array. Array of shape: [%s] was given."
                % (" ".join(("%d" % s for s in cc.shape)))
            )
        other_axis = (axis + 1) % 2
        # get pixel having the maximum value of the correlation array
        pix_max = np.argmax(cc, axis=axis)

        # select a n neighborhood for the many 1D sub-pixel fittings (with wrapping)
        p_ax_range = np.arange(-peak_radius, +peak_radius + 1)
        p_ax = (pix_max[None, :] + p_ax_range[:, None]) % img_shape[axis]

        p_ln = np.tile(np.arange(0, img_shape[other_axis])[None, :], [2 * peak_radius + 1, 1])

        # extract the pixel coordinates along the axis
        fc_ax = None if cc_coords is None else cc_coords[p_ax.flatten()].reshape(p_ax.shape)

        # extract the correlation values
        if other_axis == 0:
            f_vals = cc[p_ln, p_ax]
        else:
            f_vals = cc[p_ax, p_ln]

        return (f_vals, fc_ax)

    def _determine_roi(self, img_shape, roi_yxhw):
        if roi_yxhw is None:
            # vertical and horizontal window sizes are reduced to a power of 2
            # to accelerate fft if requested. Default is not.
            roi_yxhw = previouspow2(img_shape)
            if not self.truncate_vert_pow2:
                roi_yxhw[0] = img_shape[0]
            if not self.truncate_horz_pow2:
                roi_yxhw[1] = img_shape[1]

        roi_yxhw = np.array(roi_yxhw, dtype=np.intp)
        if len(roi_yxhw) == 2:  # Convert centered 2-element roi into 4-element
            roi_yxhw = np.concatenate(((img_shape - roi_yxhw) // 2, roi_yxhw))
        return roi_yxhw

    def _prepare_image(
        self, img, invalid_val=1e-5, roi_yxhw=None, median_filt_shape=None, low_pass=None, high_pass=None,
    ):
        """
        Prepare and returns  a cropped  and filtered image, or array of filtered images if the input is an  array of images.

        Parameters
        ----------
        img: numpy.ndarray
            image or stack of images
        invalid_val: float
            value to be used in replacement of nan and inf values
        median_filt_shape: int or sequence of int
            the width or the widths of the median window
        low_pass: float or sequence of two floats
            Low-pass filter properties, as described in `nabu.misc.fourier_filters`
        high_pass: float or sequence of two floats
            High-pass filter properties, as described in `nabu.misc.fourier_filters`

        Returns
        -------
        numpy.array_like
            The computed filter
        """
        img = np.squeeze(img)  # Removes singleton dimensions, but does a shallow copy
        img = np.ascontiguousarray(img, dtype=self.data_type)

        if roi_yxhw is not None:
            img = img[
                ..., roi_yxhw[0] : roi_yxhw[0] + roi_yxhw[2], roi_yxhw[1] : roi_yxhw[1] + roi_yxhw[3],
            ]

        img = img.copy()

        img[np.isnan(img)] = invalid_val
        img[np.isinf(img)] = invalid_val

        if high_pass is not None or low_pass is not None:
            img_filter = fourier_filters.get_bandpass_filter(
                img.shape[-2:],
                cutoff_lowpass=low_pass,
                cutoff_highpass=high_pass,
                use_rfft=True,
                data_type=self.data_type,
            )
            # fft2 and iff2 use axes=(-2, -1) by default
            img = local_ifftn(local_fftn(img, axes=(-2, -1)) * img_filter, axes=(-2, -1)).real

        if median_filt_shape is not None:
            img_shape = img.shape
            # expanding filter shape with ones, to cover the stack of images
            # but disabling inter-image filtering
            median_filt_shape = np.concatenate(
                (np.ones((len(img_shape) - len(median_filt_shape),), dtype=np.intp), median_filt_shape,)
            )
            img = median_filter(img, size=median_filt_shape)
        return img

    def _transform_to_fft(
        self, img_1: np.ndarray, img_2: np.ndarray, padding_mode, axes=(-2, -1), low_pass=None, high_pass=None
    ):
        do_circular_conv = padding_mode is None or padding_mode == "wrap"
        img_shape = img_2.shape
        if not do_circular_conv:
            pad_size = np.ceil(np.array(img_shape) / 2).astype(np.intp)
            pad_array = [(0,)] * len(img_shape)
            for a in axes:
                pad_array[a] = (pad_size[a],)

            img_1 = np.pad(img_1, pad_array, mode=padding_mode)
            img_2 = np.pad(img_2, pad_array, mode=padding_mode)
        else:
            pad_size = None
        img_shape = img_2.shape

        # compute fft's of the 2 images
        img_fft_1 = local_fftn(img_1, axes=axes)
        img_fft_2 = local_fftn(img_2, axes=axes)

        if low_pass is not None or high_pass is not None:
            filt = fourier_filters.get_bandpass_filter(
                img_shape[-2:],
                cutoff_lowpass=low_pass,
                cutoff_highpass=high_pass,
                use_rfft=True,
                data_type=self.data_type,
            )
        else:
            filt = None

        return img_fft_1, img_fft_2, filt, pad_size

    def _compute_correlation_fft(
        self, img_1: np.ndarray, img_2: np.ndarray, padding_mode, axes=(-2, -1), low_pass=None, high_pass=None
    ):
        img_fft_1, img_fft_2, filt, pad_size = self._transform_to_fft(
            img_1, img_2, padding_mode=padding_mode, axes=axes, low_pass=low_pass, high_pass=high_pass
        )

        img_prod = img_fft_1 * np.conjugate(img_fft_2)

        if filt is not None:
            img_prod *= filt

        # inverse fft of the product to get cross_correlation of the 2 images
        cc = np.real(local_ifftn(img_prod, axes=axes))

        if pad_size is not None:
            cc_shape = cc.shape
            cc = np.fft.fftshift(cc, axes=axes)

            slicing = [slice(None)] * len(cc_shape)
            for a in axes:
                slicing[a] = slice(pad_size[a], cc_shape[a] - pad_size[a])
            cc = cc[tuple(slicing)]

            cc = np.fft.ifftshift(cc, axes=axes)

        return cc

    def _add_plot_window(self, fig, ax=None):
        self._plot_windows[fig.number] = {
            "figure": fig,
            "axes": ax
        }

    def close_plot_window(self, n, errors="raise"):
        """
        Close a plot window. Applicable only if the class was instantiated with verbose=True.

        Parameters
        ----------
        n: int
            Figure number to close
        errors: str, optional
            What to do with errors. It can be either "raise", "log" or "ignore".
        """
        if not self.verbose:
            return
        if n not in self._plot_windows:
            msg = "Cannot close plot window number %d: no such window" % n
            if errors == "raise":
                raise ValueError(msg)
            elif errors == "log":
                self.logger.error(msg)
        fig_ax = self._plot_windows.pop(n)
        plt.close(fig_ax["figure"].number) # would also work with the object itself

    def close_last_plot_windows(self, n=1):
        """
        Close the last "n" plot windows.
        Applicable only if the class was instanciated with verbose=True.

        Parameters
        -----------
        n: int, optional
            Integer indicating how many plot windows should be closed.
        """
        figs_nums = sorted(self._plot_windows.keys(), reverse=True)
        n = min(n, len(figs_nums))
        for i in range(n):
            self.close_plot_window(figs_nums[i], errors="ignore")



import os
from ..resources.logger import LoggerOrPrint
from ..utils import deprecation_warning, subsample_dict

class DatasetValidatorBase:
    def __init__(self, nabu_config, dataset_info, logger=None):
        """
        Perform a coupled validation of nabu configuration against dataset information.
        Check the consistency of these two structures, and modify them in-place.

        Parameters
        ----------
        nabu_config: dict
            Dictionary containing the nabu configuration, usually got from
            `nabu.pipeline.config.validate_config()`
            It will be modified !
        dataset_info: `DatasetAnalyzer` instance
            Structure containing information on the dataset to process.
            It will be modified !
        """
        self.nabu_config = nabu_config
        self.dataset_info = dataset_info
        self.logger = LoggerOrPrint(logger)
        self._validate()


    def _validate(self):
        raise ValueError("Base class")


    # COMPAT.
    @property
    def dataset_infos(self):
        deprecation_warning(
            "The name 'dataset_infos' is deprecated. Please use 'dataset_info'.",
            func_name="validator_dataset_infos_attr"
        )
        return self.dataset_info
    # ---

    @property
    def is_halftomo(self):
        do_halftomo = self.nabu_config["reconstruction"].get("enable_halftomo", False)
        if do_halftomo == "auto":
            do_halftomo = self.dataset_info.is_halftomo
            if do_halftomo is None:
                raise ValueError(
                    "'enable_halftomo' was set to 'auto' but unable to get the information on field of view"
                )
        return do_halftomo


    def _check_not_empty(self):
        if len(self.dataset_info.projections) == 0:
            msg = "Dataset seems to be empty (no projections)"
            self.logger.fatal(msg)
            raise ValueError(msg)
        if self.dataset_info.n_angles is None:
            msg = "Could not determine the number of projections. Please check the .info or HDF5 file"
            self.logger.fatal(msg)
            raise ValueError(msg)
        for dim_name, n in zip(["dim_1", "dim_2"], self.dataset_info.radio_dims):
            if n is None:
                msg = "Could not determine %s. Please check the .info file or HDF5 file" % dim_name
                self.logger.fatal(msg)
                raise ValueError(msg)
        self._projs_indices = sorted(self.dataset_info.projections.keys())


    @staticmethod
    def _convert_negative_idx(idx, last_idx):
        res = idx
        if idx < 0:
            res = last_idx + idx
        return res


    def _get_nx_ny(self, binning=1):
        if self.is_halftomo:
            cor = int(round(self.dataset_info.axis_position / binning))
            nx = self.dataset_info.radio_dims[0] // binning
            nx = max(2*cor, 2 * (nx - 1 - cor))
        else:
            nx = self.dataset_info.radio_dims[0] // binning
        ny = nx
        return nx, ny


    def _convert_negative_indices(self):
        """
        Convert any negative index to the corresponding positive index.
        """
        nx, nz = self.dataset_info.radio_dims
        ny = nx
        if self.is_halftomo:
            if self.dataset_info.axis_position is None:
                raise ValueError("Cannot use rotation axis position in the middle of the detector when half tomo is enabled")
            nx, ny = self._get_nx_ny()
        what = (
            ("reconstruction", "start_x", nx),
            ("reconstruction", "end_x", nx),
            ("reconstruction", "start_y", ny),
            ("reconstruction", "end_y", ny),
            ("reconstruction", "start_z", nz),
            ("reconstruction", "end_z", nz),
        )
        for section, key, upper_bound in what:
            val = self.nabu_config[section][key]
            if isinstance(val, str):
                idx_mapping = {
                    "first": 0,
                    "middle": upper_bound // 2, # works on both start_ and end_ since the end_ index is included
                    "last": upper_bound - 1
                }
                res = idx_mapping[val]
            else:
                res = self._convert_negative_idx(
                self.nabu_config[section][key], upper_bound
            )
            self.nabu_config[section][key] = res


    def _get_resources(self):
        opts = self.nabu_config["resources"]
        if opts["gpu_id"] != []:
            opts["gpus"] = len(opts["gpu_id"])
        if opts["gpus"] == 0:
            opts["gpu_id"] = []


    def _get_output_filename(self):
        opts = self.nabu_config["output"]
        dataset_path = self.nabu_config["dataset"]["location"]
        if opts["location"] == "" or opts["location"] is None:
            opts["location"] = os.path.dirname(dataset_path)
        if opts["file_prefix"] == "" or opts["file_prefix"] is None:
            if os.path.isfile(dataset_path): # hdf5
                file_prefix = os.path.basename(dataset_path).split(".")[0]
            elif os.path.isdir(dataset_path):
                file_prefix = os.path.basename(dataset_path)
            else:
                raise ValueError(
                    "dataset location %s is neither a file or directory"
                    % dataset_path
                )
            file_prefix += "_rec" # avoid overwriting dataset
            opts["file_prefix"] = file_prefix


    @staticmethod
    def _check_start_end_idx(start, end, n_elements, start_name="start_x", end_name="end_x"):
        assert (start >= 0 and start < n_elements), "Invalid value for %s, must be >= 0 and < %d" % (start_name, n_elements)
        assert (end >= 0 and end < n_elements), "Invalid value for %s, must be >= 0 and < %d" % (end_name, n_elements)


    # COMPAT.
    def perform_all_checks(self, remove_unused_radios=True):
        """
        Deprecated function
        """
        deprecation_warning(
            "This method is deprecated as validation is now done at class instantiation",
            func_name="perform_all_checks"
        )
    # ----


    def _handle_binning(self):
        """
        Modify the dataset description and nabu config to handle binning and
        projections subsampling.
        """
        self.dataset_info._radio_dims_notbinned = self.dataset_info.radio_dims
        dataset_cfg = self.nabu_config["dataset"]
        self.binning = (dataset_cfg["binning"], dataset_cfg["binning_z"])
        self.dataset_info._binning = self.binning
        subsampling_factor = dataset_cfg["projections_subsampling"]
        self.projections_subsampling = subsampling_factor
        self.dataset_info._projections_subsampled = self.dataset_info.projections
        self.dataset_info._projs_indices_subsampled = self._projs_indices
        if subsampling_factor > 1:
            self.dataset_info._projections_subsampled = subsample_dict(self.dataset_info.projections, subsampling_factor)
            self.dataset_info._projs_indices_subsampled = sorted(self.dataset_info._projections_subsampled.keys())
            self.dataset_info.reconstruction_angles = self.dataset_info.reconstruction_angles[::subsampling_factor]
            # should be simply len(projections)... ?
            self.dataset_info.n_angles //= subsampling_factor
        if self.binning != (1, 1):
            bin_x, bin_z = self.binning
            bin_y = bin_x # square slices
            # Update end_x, end_y, end_z
            rec_cfg = self.nabu_config["reconstruction"]
            end_x, end_y = self._get_end_xy() # Not so trivial. See function documentation
            rec_cfg["end_x"] = end_x
            rec_cfg["end_y"] = end_y
            rec_cfg["end_z"] = (rec_cfg["end_z"] + 1) // bin_z - 1
            rec_cfg["start_x"] = (rec_cfg["start_x"] // bin_x)
            rec_cfg["start_y"] = (rec_cfg["start_y"] // bin_y)
            rec_cfg["start_z"] = (rec_cfg["start_z"] // bin_z)
            # Update radio_dims
            nx, nz = self.dataset_info.radio_dims
            nx //= bin_x
            ny = nx # square slices
            nz //= bin_z
            self.dataset_info.radio_dims = (nx, nz)
            # Update the Rotation center
            # TODO axis_corrections
            rot_c = self.dataset_info.axis_position
            nx0, nz0 = self.dataset_info._radio_dims_notbinned
            bin_fact = nx0 / nx
            if rot_c is not None: # user-specified
                rot_c /= bin_fact # float
            else:
                rot_c = (nx - 1)/2.
            self.dataset_info.axis_position = rot_c


    def _check_output_file(self):
        out_cfg = self.nabu_config["output"]
        out_fname = os.path.join(out_cfg["location"], out_cfg["file_prefix"] + out_cfg["file_format"])
        if os.path.exists(out_fname):
            raise ValueError("File %s already exists" % out_fname)


    def _handle_processing_mode(self):
        mode = self.nabu_config["resources"]["method"]
        if mode == "preview":
            print("Warning: the method 'preview' was selected. This means that the data volume will be binned so that everything fits in memory.")
            # TODO automatically compute binning/subsampling factors as a function of lowest memory (GPU)
            self.nabu_config["dataset"]["binning"] = 2
            self.nabu_config["dataset"]["binning_z"] = 2
            self.nabu_config["dataset"]["projections_subsampling"] = 2
        # TODO handle other modes


    def _get_end_xy(self):
        """
        Return the "end_x" value for reconstruction, accounting for binning.

        There are three situations:

           1. Normal setting: Nx_rec = Nx
           2. Half acquisition, CoR on the right side: Nx_rec = 2 * |c|
           3. Half acquisition, CoR on the left side: Nx_rec = 2 * (Nx - 1 - |c|)

        where |c| = int(round(c)).

        **Without binnig**

        Let e0 denote the default value for "end_x", without user modification.
        By default, all the slice is reconstructed, so
            e0 = Nx_rec - 1
        Let e denote the user value for "end_x". By default e = e0, but the user might
        want to tune it.
        Let d denote the distance between e and e0: d = e0 - e. By default d = 0.

        **With binning**

        Let b denote the binning value in x.

           1. Nx' = Nx // b
           2. Nx' = 2 * |c/b|
           3. Nx' = 2 * (Nx//b - 1 - |c/b|)

        With the same notations, with a prime suffix, we have:

           * e0' = Nx' - 1  is the default value of "end_x" accounting for binning
           * e' is the user value for "end_x" accounting for binning
           * d' = e0' - e'  is the distance between e0' and e'

        In the standard setting (no half tomography), computing e' (i.e end_x) is straightforward:
        e' = (e+1)//b - 1

        With half tomography, because of the presence of |c| = int(floor(c)), the things
        are a bit more difficult.
        We enforce the following invariant in all settings:

           (I1) dist(e0', e') = dist(e0, e) // b

        Another possible invariant is

            (I2) delta_x' = delta_x // b

        Which is equivalent to (I1) only in setting (1) (i.e no half-tomography).
        In the half-tomography setting, (I1) and (I2) are not equivalent anymore, so we have
        to choose between the two.
        We believe it is more consistent to preserve "end_x", so that a user not modying "end_x"
        ends up with all the range [0, Nx - 1] reconstructed.

        Therefore:
           e' = e0' - d//b

        """
        b, _ = self.binning
        end_xy = []
        for i in range(2):
            what = ["end_x", "end_y"][i]
            e0 = self._get_nx_ny()[i] - 1
            e0p = self._get_nx_ny(binning=b)[i] - 1
            d = e0 - self.nabu_config["reconstruction"][what]
            ep = e0p - d//b
            end_xy.append(ep)
        return tuple(end_xy)




import numpy as np
try:
    from silx.opencl.processing import EventDescription, OpenclProcessing
    from silx.opencl.convolution import Convolution as CLConvolution
    import pyopencl.array as parray
    from pyopencl.elementwise import ElementwiseKernel
    __have_opencl__ = True
except ImportError:
    __have_opencl__ = False
from .unsharp import UnsharpMask

class OpenclUnsharpMask(UnsharpMask, OpenclProcessing):
    def __init__(self, shape, sigma, coeff, mode="reflect", method="gaussian",
                ctx=None, devicetype="all", platformid=None, deviceid=None,
                block_size=None, memory=None, profile=False):
        """
        NB: For now, this class is designed to use the lowest amount of GPU memory
        as possible. Therefore, the input and output image/volumes are assumed
        to be already on device.
        """
        if not(__have_opencl__):
            raise ImportError("Need silx and pyopencl")
        OpenclProcessing.__init__(
            self, ctx=ctx, devicetype=devicetype,
            platformid=platformid, deviceid=deviceid, block_size=block_size,
            memory=memory, profile=profile
        )
        UnsharpMask.__init__(self, shape, sigma, coeff, mode=mode)
        self._init_convolution()
        self._init_mad_kernel()

    def _init_convolution(self):
        self.convolution = CLConvolution(
            self.shape,
            self._gaussian_kernel,
            mode=self.mode,
            ctx=self.ctx, profile=self.profile,
            extra_options={ # Use the lowest amount of memory
                "allocate_input_array": False,
                "allocate_output_array": False,
                "allocate_tmp_array": True,
                "dont_use_textures": True,
            }
        )

    def _init_mad_kernel(self):
        # parray.Array.mul_add is out of place...
        self.mad_kernel = ElementwiseKernel(
            self.ctx,
            "float* array, float fac, float* other, float otherfac",
            "array[i] = fac * array[i] + otherfac * other[i]",
            name="mul_add"
        )

    def unsharp(self, image, output):
        # For now image and output are assumed to be already allocated on device
        assert isinstance(image, parray.Array)
        assert isinstance(output, parray.Array)
        self.convolution(image, output=output)
        if self.method == "gaussian":
            self.mad_kernel(output, -self.coeff, image, 1. + self.coeff)
        else:
            self.mad_kernel(output, self.coeff, image, 1.)
        return output



import numpy as np


def rescale_data(data, new_min, new_max, data_min=None, data_max=None):
    if data_min is None:
        data_min = np.min(data)
    if data_max is None:
        data_max = np.max(data)
    return (new_max - new_min)/(data_max - data_min) * (data - data_min) + new_min


def get_dtype_range(dtype, normalize_floats=False):
    if np.dtype(dtype).kind in ["u", "i"]:
        dtype_range = (np.iinfo(dtype).min, np.iinfo(dtype).max)
    else:
        if normalize_floats:
            dtype_range = (-1., 1.)
        else:
            dtype_range = (np.finfo(dtype).min, np.finfo(dtype).max)
    return dtype_range


def psnr(img1, img2):
    if img1.dtype != img2.dtype:
        raise ValueError("both images should have the same data type")
    dtype_range = get_dtype_range(img1.dtype, normalize_floats=True)
    dtype_range = dtype_range[-1] - dtype_range[0]
    if np.dtype(img1.dtype).kind in ["f", "c"]:
        img1 = rescale_data(img1, -1., 1.)
        img2 = rescale_data(img2, -1., 1.)
    mse = np.mean((img1.astype(np.float64) - img2)**2)
    return 10 * np.log10((dtype_range ** 2) / mse)



#
# silx.opencl.utils.ConvolutionInfos cannot be used as long as
# silx.opencl instantiates the "ocl" singleton in __init__,
# leaving opencl contexts all over the place in some cases.
#
# so for now: copypasta
#

class ConvolutionInfos(object):
    allowed_axes = {
        "1D": [None],
        "separable_2D_1D_2D": [None, (0, 1), (1, 0)],
        "batched_1D_2D": [(0,), (1,)],
        "separable_3D_1D_3D": [
            None,
            (0, 1, 2),
            (1, 2, 0),
            (2, 0, 1),
            (2, 1, 0),
            (1, 0, 2),
            (0, 2, 1)
        ],
        "batched_1D_3D": [(0,), (1,), (2,)],
        "batched_separable_2D_1D_3D": [(0,), (1,), (2,)], # unsupported (?)
        "2D": [None],
        "batched_2D_3D": [(0,), (1,), (2,)],
        "separable_3D_2D_3D": [
            (1, 0),
            (0, 1),
            (2, 0),
            (0, 2),
            (1, 2),
            (2, 1),
        ],
        "3D": [None],
    }
    use_cases = {
        (1, 1): {
            "1D": {
                "name": "1D convolution on 1D data",
                "kernels": ["convol_1D_X"],
            },
        },
        (2, 2): {
            "2D": {
                "name": "2D convolution on 2D data",
                "kernels": ["convol_2D_XY"],
            },
        },
        (3, 3): {
            "3D": {
                "name": "3D convolution on 3D data",
                "kernels": ["convol_3D_XYZ"],
            },
        },
        (2, 1): {
            "separable_2D_1D_2D": {
                "name": "Separable (2D->1D) convolution on 2D data",
                "kernels": ["convol_1D_X", "convol_1D_Y"],
            },
            "batched_1D_2D": {
                "name": "Batched 1D convolution on 2D data",
                "kernels": ["convol_1D_X", "convol_1D_Y"],
            },
        },
        (3, 1): {
            "separable_3D_1D_3D": {
                "name": "Separable (3D->1D) convolution on 3D data",
                "kernels": ["convol_1D_X", "convol_1D_Y", "convol_1D_Z"],
            },
            "batched_1D_3D": {
                "name": "Batched 1D convolution on 3D data",
                "kernels": ["convol_1D_X", "convol_1D_Y", "convol_1D_Z"],
            },
            "batched_separable_2D_1D_3D": {
                "name": "Batched separable (2D->1D) convolution on 3D data",
                "kernels": ["convol_1D_X", "convol_1D_Y", "convol_1D_Z"],
            },
        },
        (3, 2): {
            "separable_3D_2D_3D": {
                "name": "Separable (3D->2D) convolution on 3D data",
                "kernels": ["convol_2D_XY", "convol_2D_XZ", "convol_2D_YZ"],
            },
            "batched_2D_3D": {
                "name": "Batched 2D convolution on 3D data",
                "kernels": ["convol_2D_XY", "convol_2D_XZ", "convol_2D_YZ"],
            },
        },
    }




# COMPAT.
from ..utils import deprecated_class
from ..reconstruction.rings import MunchDeringer as MunchDeringer_, munchetal_filter as munchetal_filter_

MunchDeringer = deprecated_class(
    "preproc.rings was moved to reconstruction.rings",
    do_print=True
)(MunchDeringer_)
munchetal_filter = deprecated_class(
    "preproc.rings was moved to reconstruction.rings",
    do_print=True
)(munchetal_filter_)



import atexit
from math import ceil
import numpy as np
from ..resources.gpu import GPUDescription
try:
    import pycuda
    import pycuda.driver as cuda
    from pycuda import gpuarray as garray
    from pycuda.tools import make_default_context
    from pycuda.tools import clear_context_caches
    __has_pycuda__ = True
    __pycuda_error_msg__ = None
    if pycuda.VERSION[0] < 2020:
        print("Error: need pycuda >= 2020.1")
        __has_pycuda__ = False
except ImportError as err:
    __has_pycuda__ = False
    __pycuda_error_msg__ = str(err)
try:
    import skcuda
    __has_cufft__ = True
except ImportError:
    __has_cufft__ = False


def get_cuda_context(device_id=None, cleanup_at_exit=True):
    """
    Create or get a CUDA context.
    """
    current_ctx = cuda.Context.get_current()
    # If a context already exists, use this one
    # TODO what if the device used is different from device_id ?
    if current_ctx is not None:
        return current_ctx
    # Otherwise create a new context
    cuda.init()

    if device_id is None:
        device_id = 0
    # Use the Context obtained by retaining the device's primary context,
    # which is the one used by the CUDA runtime API (ex. scikit-cuda).
    # Unlike Context.make_context(), the newly-created context is not made current.
    context = cuda.Device(device_id).retain_primary_context()
    context.push()
    # Register a clean-up function at exit
    def _finish_up(context):
        if context is not None:
            context.pop()
            context = None
        clear_context_caches()
    if cleanup_at_exit:
        atexit.register(_finish_up, context)
    return context


def count_cuda_devices():
    if cuda.Context.get_current() is None:
        cuda.init()
    return cuda.Device.count()


def get_gpu_memory(device_id):
    """
    Return the total memory (in GigaBytes) of a device.
    """
    cuda.init()
    return cuda.Device(device_id).total_memory()/1e9


def is_gpu_usable():
    """
    Test whether at least one Nvidia GPU is available.
    """
    try:
        n_gpus = count_cuda_devices()
    except Exception as exc:
        # Fragile
        if exc.__str__() != "cuInit failed: no CUDA-capable device is detected":
            raise
        n_gpus = 0
    res = (n_gpus > 0)
    return res


def detect_cuda_gpus():
    """
    Detect the available Nvidia CUDA GPUs on the current host.

    Returns
    --------
    gpus: dict
        Dictionary where the key is the GPU ID, and the value is a `pycuda.driver.Device` object.
    error_msg: str
        In the case where there is an error, the message is returned in this item.
        Otherwise, it is a None object.
    """
    gpus = {}
    error_msg = None
    if not(__has_pycuda__):
        return {}, __pycuda_error_msg__
    try:
        cuda.init()
    except Exception as exc:
        error_msg = str(exc)
    if error_msg is not None:
        return {}, error_msg
    try:
        n_gpus = cuda.Device.count()
    except Exception as exc:
        error_msg = str(exc)
    if error_msg is not None:
        return {}, error_msg
    for i in range(n_gpus):
        gpus[i] = cuda.Device(i)
    return gpus, None


def collect_cuda_gpus():
    """
    Return a dictionary of GPU ids and brief description of each CUDA-compatible
    GPU with a few fields.
    """
    gpus, error_msg = detect_cuda_gpus()
    if error_msg is not None:
        return None
    cuda_gpus = {}
    for gpu_id, gpu in gpus.items():
        cuda_gpus[gpu_id] = GPUDescription(gpu).get_dict()
    return cuda_gpus


"""
pycuda/driver.py
    np.complex64: SIGNED_INT32, num_channels = 2
    np.float64:  SIGNED_INT32, num_channels = 2
    np.complex128: array_format.SIGNED_INT32, num_channels = 4

double precision: pycuda-helpers.hpp:
  typedef float fp_tex_float;   // --> float32
  typedef int2 fp_tex_double;   // --> float64
  typedef uint2 fp_tex_cfloat;  // --> complex64
  typedef int4 fp_tex_cdouble;  // --> complex128
"""
def cuarray_format_to_dtype(cuarr_fmt):
    # reverse of cuda.dtype_to_array_format
    fmt = cuda.array_format
    mapping = {
        fmt.UNSIGNED_INT8: np.uint8,
        fmt.UNSIGNED_INT16: np.uint16,
        fmt.UNSIGNED_INT32: np.uint32,
        fmt.SIGNED_INT8: np.int8,
        fmt.SIGNED_INT16: np.int16,
        fmt.SIGNED_INT32: np.int32,
        fmt.FLOAT: np.float32,
    }
    if cuarr_fmt not in mapping:
        raise TypeError("Unknown format %s" % cuarr_fmt)
    return mapping[cuarr_fmt]


def cuarray_shape_dtype(cuarray):
    desc = cuarray.get_descriptor_3d()
    shape = (desc.height, desc.width)
    if desc.depth > 0:
        shape = (desc.depth, ) + shape
    dtype = cuarray_format_to_dtype(desc.format)
    return shape, dtype


def get_shape_dtype(arr):
    if isinstance(arr, garray.GPUArray) or isinstance(arr, np.ndarray):
        return arr.shape, arr.dtype
    elif isinstance(arr, cuda.Array):
        return cuarray_shape_dtype(arr)
    else:
        raise ValueError("Unknown array type %s" % str(type(arr)))


def copy_array(dst, src, check=False, src_dtype=None):
    """
    Copy a source array to a destination array.
    Source and destination can be either numpy.ndarray, pycuda.Driver.Array,
    or pycuda.gpuarray.GPUArray.

    Parameters
    -----------
    dst: pycuda.driver.Array or pycuda.gpuarray.GPUArray or numpy.ndarray
        Destination array. Its content will be overwritten by copy.
    src: pycuda.driver.Array or pycuda.gpuarray.GPUArray or numpy.ndarray
        Source array.
    check: bool, optional
        Whether to check src and dst shape and data type.
    """

    shape_src, dtype_src = get_shape_dtype(src)
    shape_dst, dtype_dst = get_shape_dtype(dst)
    dtype_src = src_dtype or dtype_src
    if check:
        if shape_src != shape_dst:
            raise ValueError("shape_src != shape_dst : have %s and %s" % (str(shape_src), str(shape_dst)))
        if dtype_src != dtype_dst:
            raise ValueError("dtype_src != dtype_dst : have %s and %s" % (str(dtype_src), str(dtype_dst)))
    if len(shape_src) == 2:
        copy = cuda.Memcpy2D()
        h, w = shape_src
    elif len(shape_src) == 3:
        copy = cuda.Memcpy3D()
        d, h, w = shape_src
        copy.depth = d
    else:
        raise ValueError("Expected arrays with 2 or 3 dimensions")

    if isinstance(src, cuda.Array):
        copy.set_src_array(src)
    elif isinstance(src, garray.GPUArray):
        copy.set_src_device(src.gpudata)
    else: # numpy
        copy.set_src_host(src)

    if isinstance(dst, cuda.Array):
        copy.set_dst_array(dst)
    elif isinstance(dst, garray.GPUArray):
        copy.set_dst_device(dst.gpudata)
    else: # numpy
        copy.set_dst_host(dst)

    copy.width_in_bytes = copy.dst_pitch = w * np.dtype(dtype_src).itemsize
    copy.dst_height = copy.height = h
    # ??
    if len(shape_src) == 2:
        copy(True)
    else:
        copy()
    ###



def copy_big_gpuarray(dst, src, itemsize=4, checks=False):
    """
    Copy a big `pycuda.gpuarray.GPUArray` into another.
    Transactions of more than 2**32 -1 octets fail, so are doing several
    partial copies of smaller arrays.
    """
    d2h = isinstance(dst, np.ndarray)
    if checks:
        assert dst.dtype == src.dtype
        assert dst.shape == src.shape
    limit = 2**32 - 1
    if np.prod(dst.shape) * itemsize < limit:
        if d2h:
            src.get(ary=dst)
        else:
            dst[:] = src[:]
        return
    def get_shape2(shape):
        shape2 = list(shape)
        while np.prod(shape2)*4 > limit:
            shape2[0] //= 2
        return tuple(shape2)
    shape2 = get_shape2(dst.shape)
    nz0 = dst.shape[0]
    nz = shape2[0]
    n_transfers = ceil(nz0 / nz)
    for i in range(n_transfers):
        zmax = min((i+1)*nz, nz0)
        if d2h:
            src[i*nz:zmax].get(ary=dst[i*nz:zmax])
        else:
            dst[i*nz:zmax] = src[i*nz:zmax]


def replace_array_memory(arr, new_shape):
    """
    Replace the underlying buffer data of a  `pycuda.gpuarray.GPUArray`.
    This function is dangerous !
    It should merely be used to clear memory, the array should not be used afterwise.
    """
    arr.gpudata.free()
    arr.gpudata = arr.allocator(int(np.prod(new_shape) * arr.dtype.itemsize))
    arr.shape = new_shape
    # TODO re-compute strides
    return arr




"""
This file is obsolete. Don't rely on it.
"""

from math import ceil
from socket import gethostname
from multiprocessing import cpu_count

from distributed import Client, get_worker, worker_client, LocalCluster, Nanny, SpecCluster, Scheduler

from ..utils import check_supported
from ..resources.computations import estimate_chunk_size
from ..resources.processconfig import ProcessConfig
from ..resources.gpu import pick_gpus
from ..resources.utils import get_memory_per_node, get_threads_per_node
from ..cuda.utils import collect_cuda_gpus
from ..opencl.utils import collect_opencl_gpus, collect_opencl_cpus, pick_opencl_cpu_platform
from ..app.logger import Logger, LoggerOrPrint
from ..app.process import WorkerProcess


def get_dataset(dataset_name, client=None):
    if client is None:
        with worker_client() as client:
            res = client.datasets[dataset_name]
    else:
        res = client.datasets[dataset_name]
    return res


def get_dask_worker():
    try:
        w = get_worker()
    except Exception as exc:
        w = None
    return w


def get_worker_resources(try_cuda=True, try_opencl=True):
    current_worker = get_worker()
    vm = virtual_memory()
    resources = {
        "worker_name": current_worker.name,
        "worker_addr": current_worker.address,
        "host": gethostname(),
        "mem_total_GB": vm.total / 1e9,
        "mem_avail_GB": vm.available / 1e9,
        "cpu_cores": cpu_count(),
        "cuda_gpus": {},
        "opencl_gpus": {},
    }
    if try_cuda:
        resources["cuda_gpus"] = collect_cuda_gpus()
    if try_opencl:
        resources["opencl_gpus"] = collect_opencl_gpus()
    return resources


def get_workers_resources(client, try_cuda=True, try_opencl=True):
    workers = list(client.has_what().keys())
    resources = {}
    for worker_addr in workers:
        f = client.submit(
            get_worker_resources, try_cuda=try_cuda, try_opencl=try_opencl,
            workers=[worker_addr], pure=False
        )
        resources[worker_addr] = f.result()
    return resources


def get_gpu_workers(workers_resources, max_gpu_per_worker=1):
    """
    Get the workers that can use a GPU.

    Parameters
    ----------
    workers_resources: dict
        Dictionary of workers resources obtained with `get_workers_resources`
    max_gpu_per_worker: int, optional
        Maximum GPUs allower per worker. If a worker can "see" more than
        `max_gpu_per_worker`, it will raise an error.
    Returns
    -------
    gpu_workers: dict
        Dictionary of workers that can use a GPU, along with the usable GPUs.
    nongpu_workers: list
        List of workers that cannot use a GPU
    """
    gpu_workers = {}
    nongpu_workers = []
    for worker_name, resources in workers_resources.items():
        w_gpus = {"cuda_gpus": {}, "opencl_gpus": {}}
        for gpu_type in ["cuda_gpus", "opencl_gpus"]:
            if gpu_type not in resources:
                continue
            gpus = resources[gpu_type] or {}
            if len(gpus) > max_gpu_per_worker:
                raise ValueError(
                    "Expected at most %d GPUs, got %d"
                    % (max_gpu_per_worker, len(gpus))
                )
            if len(gpus) == 0:
                continue
            w_gpus[gpu_type] = gpus
        if len(w_gpus["opencl_gpus"]) == 0 and len(w_gpus["cuda_gpus"]) == 0:
            nongpu_workers.append(worker_name)
        else:
            gpu_workers[worker_name] = w_gpus
    return gpu_workers, nongpu_workers




# TODO callbacks ? Several options:
#  - one callback after processing a chunk (i.e after this function)
#    either the worker or the client can do the callback (although data is on the worker side)
#  - one callback after each individual sub-processing (i.e phase retrieval, etc)
#    in this case, this should be registered in processing_options.
#    The callback is done on the worker side.
def worker_process_chunk(
    sub_region, chunk_size, chunk_id,
    logfile=None, loglevel="debug", extra_options=None,
    destroy_workerprocess_class=True,
):
    """
    Entry point for a `dask.distributed` `Worker`.

    Parameters
    -----------
    sub_region: tuple
        Tuple describing the region to process in the data volume, in the form
        `(start_x, end_x, start_z, end_z)`.
    chunk_size: int
        Size of the chunk ("delta z")
    chunk_id: int
        Index of the current chunk.
    logfile: str, optional
        Name of the log file.
    extra_options: dict
        Dictionary of extra options for `WorkerProcess`.
    """
    worker = get_worker()

    worker_conf = worker.client.datasets[worker.name]
    process_config = worker.client.datasets["process_config"]

    logname = "nabu-%s" % worker.name
    if logfile is None:
        logfile = logname + ".log"
    logger = Logger(logname, level=loglevel, logfile=logfile)

    ##
    extra_options = extra_options or {}
    extra_options["clear_gpu_memory_after_buildsino"] = True
    ##

    logger.debug("%s: spawning WorkerProcess" % worker.name)
    W = WorkerProcess(
        process_config, sub_region, chunk_size=chunk_size,
        use_cuda=worker_conf["use_cuda"], use_opencl=worker_conf["use_opencl"],
        logger=logger, extra_options=extra_options
    )
    logger.debug("%s: start processing chunk" % worker.name)
    W.process_chunk()
    # DEBUG
    if destroy_workerprocess_class:
        W._destroy_gpu_context()
        del W
        import gc
        gc.collect()
    #







def actor_process_chunk(sub_region, chunk_id):
    w = get_worker()
    worker_process = w.actors[list(w.actors.keys())[0]]
    worker_process.logger.info("Will process subregion %s" % str(sub_region))
    worker_process.process_chunk(sub_region=sub_region)










def _get_n_cpu_workers(n_cpu_workers, n_gpu_workers):
    if n_cpu_workers < 0:
        # (Poor) convention: total number of CPU workers
        return -n_cpu_workers
    # "X cpus workers by gpu worker
    return n_cpu_workers * n_gpu_workers




class WorkersManager:
    """
    A class for managing "Nabu workers". It has several purposes:
      - Get the requested computing resources from user-provided configuration
      - Get the available computing resources
      - Spwan the workers (either local or distributed)
      - Distribute the work to be done among workers (here chunks or images)
    """

    def __init__(self, process_config, logger=None, extra_options=None):
        """
        Initialize a WorkersManager.

        Parameters
        ----------
        process_config: `nabu.resources.processing.ProcessConfig`
            Structure describing the user parameters and dataset.
        logger: `logging.logger` or `nabu.app.logger.Logger`, optional
            Logging object.
        extra_options: dict, optional
            Dictionary of advanced options. Current options are:
               - gpu_pick_method: "cuda" or "auto"
               - chunk_size_method: "same" or "cpugpu"
               - max_chunk_size: int
        """
        self.process_config = process_config
        self.logger = LoggerOrPrint(logger)
        self._set_extra_options(extra_options)
        self._get_requested_resources()


    def _set_extra_options(self, extra_options):
        if extra_options is None:
            extra_options = {}
        advanced_options = {
            "gpu_pick_method": "cuda",
            "chunk_size_method": "same",
            "max_chunk_size": None,
        }
        advanced_options.update(extra_options)
        self.extra_options = advanced_options
        self._max_chunk_size = self.extra_options["max_chunk_size"]


    def _get_requested_resources(self):
        resources_cfg = self.process_config.nabu_config["resources"]
        self.distribution_method = resources_cfg["method"]
        #
        if self.distribution_method == "slurm":
            raise NotImplementedError()
        #
        gpu_ids = resources_cfg["gpu_id"]
        n_gpu_workers = resources_cfg["gpus"]
        n_cpu_workers = _get_n_cpu_workers(resources_cfg["cpu_workers"], n_gpu_workers)
        if (n_cpu_workers == 0) and (n_gpu_workers == 0):
            raise ValueError("Got 0 CPU workers and 0 GPU workers. Need at least one worker.")
        self.n_gpu_workers = n_gpu_workers
        self.n_cpu_workers = n_cpu_workers


    def _configure_chunk_size_same(self):
        """
        Configure the chunk size used by workers.
        In this approach, the same chunk size is used for all workers.
           Pros: simpler, especially when it comes to distribute work with client.submit()
           Cons: if max_available_chunk_size differs too much between workers, might be inefficient
        """
        chunk_sizes = []
        for worker_name, worker_desc in self.workers.items():
            chunk_sizes.append(worker_desc["chunk_size"])
        self._common_chunk_size = min(chunk_sizes)
        if self._max_chunk_size is not None:
            self._common_chunk_size = min(self._max_chunk_size, chunk_sizes)
        self._chunk_sizes = dict.fromkeys(self.workers.keys(), self._common_chunk_size)


    def _configure_chunk_size_cpu_gpu(self, gpu_cpu_ratio=2):
        """
        Configure the chunk size used by workers.
        In this approach, we use a different chunk size for GPU workers and CPU
        workers. The rationale is that GPU workers are likely to process their
        chunk much faster, so we decrease the chunk size of CPU worker by a certain
        factor in order to re-equilibrate workload.
        """
        res = {}
        for worker_name, worker_desc in self.workers.items():
            chunk_size = worker_desc["chunk_size"]
            if worker_desc["type"] == "CPU":
                chunk_size = chunk_size / gpu_cpu_ratio
            res[worker_name] = chunk_size
        self._chunk_sizes = res


    def _configure_chunk_size(self):
        method = self.extra_options["chunk_size_method"]
        if method == "same":
            self._configure_chunk_size_same()
        elif method == "cpugpu":
            self._configure_chunk_size_cpu_gpu()
        else:
            raise ValueError("Unknown chunk size method")



class LocalWorkersManager(WorkersManager):

    def __init__(self, process_config, logger=None, extra_options=None, scheduler_kwargs=None):
        """
        Initialize a LocalWorkersManager.

        Parameters
        ----------
        process_config: `nabu.resources.processing.ProcessConfig`
            Structure describing the user parameters and dataset.
        logger: `logging.logger` or `nabu.app.logger.Logger`, optional
            Logging object.
        extra_options: dict, optional
            Dictionary of advanced options. Current options are:
               - gpu_pick_method: "cuda" or "auto"
        scheduler_kwargs: dict, optional
            Extra options to pass to `distributed.Scheduler`.
        """
        super().__init__(process_config, logger=logger, extra_options=extra_options)
        self.scheduler_kwargs = scheduler_kwargs or {}
        self._configure_worker_resources()
        self._spawn_workers()
        self._estimate_workers_chunk_size()
        self._configure_workers_options()
        self._create_tasks()
        self._spawn_workers_pipelines()


    def _configure_worker_resources(self):
        resources_cfg = self.process_config.nabu_config["resources"]
        # Get memory per worker
        mem_per_node = resources_cfg["memory_per_node"]
        memory = get_memory_per_node(mem_per_node[0], is_percentage=mem_per_node[1]) * 1e9
        # Get threads per node
        threads_per_node = resources_cfg["threads_per_node"]
        threads = get_threads_per_node(threads_per_node[0], is_percentage=threads_per_node[1])
        # Pick GPUs
        self.gpus = pick_gpus(
            self.extra_options["gpu_pick_method"],
            collect_cuda_gpus(),
            collect_opencl_gpus(),
            self.n_gpu_workers
        )
        # Pick CPU
        self.cpu = None
        if self.n_cpu_workers > 0:
            self.cpu = pick_opencl_cpu_platform(collect_opencl_cpus())
        # GPU workers specification
        n_workers = self.n_cpu_workers + self.n_gpu_workers
        gpu_workers_spec = {}
        workers = {}
        mem_used_by_gpu_workers = 0
        for i in range(self.n_gpu_workers):
            mem_limit = min(memory/n_workers, self.gpus[i]["memory_GB"] * 1e9 * 2)
            mem_used_by_gpu_workers += mem_limit
            worker_name = "GPU-worker-%02d" % i
            gpu_workers_spec[i] = {
                "cls": Nanny,
                "options": {
                    "nthreads": 1, #2,
                    "name": worker_name,
                    "memory_limit": mem_limit,
                },
            }
            workers[worker_name] = {
                "name": worker_name,
                "CPU_memory_GB": mem_limit / 1e9,
                "GPU": self.gpus[i],
            }
        # CPU workers specification
        cpu_workers_spec = {}
        for i in range(self.n_cpu_workers):
            mem_limit = (memory - mem_used_by_gpu_workers)/n_workers
            worker_name = "CPU-worker-%02d" % i
            cpu_workers_spec[self.n_gpu_workers + i] = {
                "cls": Nanny,
                "options": {
                    "nthreads": int((threads - 2*self.n_gpu_workers)/self.n_cpu_workers),
                    "name": worker_name,
                    "memory_limit": mem_limit,
                },
            }
            workers[worker_name] = {
                "name": worker_name,
                "CPU_memory_GB": mem_limit / 1e9
            }
        self.workers = workers
        self._gpu_workers_spec = gpu_workers_spec
        self._cpu_workers_spec = cpu_workers_spec
        self._workers_spec = {}
        self._workers_spec.update(self._gpu_workers_spec)
        self._workers_spec.update(self._cpu_workers_spec)


    def _spawn_workers(self):
        self.logger.debug("Creating SpecCluster()")
        self.scheduler_spec = {
            "cls": Scheduler,
            "options": self.scheduler_kwargs
        }
        self.cluster = SpecCluster(
            scheduler=self.scheduler_spec,
            workers=self._workers_spec
         )
        self.client = Client(self.cluster.scheduler_address)
        self._get_workers_addresses()


    def _get_workers_addresses(self):
        def get_worker_name_type():
             curr_worker = get_worker()
             worker_type = "GPU" if "GPU" in curr_worker.name else "CPU"
             return curr_worker.name, worker_type
        for worker_addr in self.client.has_what().keys():
            f = self.client.submit(get_worker_name_type, workers=[worker_addr], pure=False)
            w_name, w_type = f.result()
            self.workers[w_name].update({"type": w_type, "address": worker_addr})


    def _estimate_workers_chunk_size(self):
        self.logger.debug("Estimating workers chunk size")
        chunks = {}
        for worker_name in self.workers.keys():
            worker_resources = self.workers[worker_name]
            mem = worker_resources["CPU_memory_GB"]
            if self.workers[worker_name]["type"] == "GPU":
                mem = min(mem, worker_resources["GPU"]["memory_GB"]*0.95)
            self.workers[worker_name]["chunk_size"] = estimate_chunk_size(
                mem, self.process_config, chunk_step=20
            )


    def _configure_workers_options(self):
        self.client.datasets.clear() # not mandatory
        workers_conf = {"process_config": self.process_config}
        # Enable/Disable cuda for GPU/CPU workers
        for worker_name, worker_desc in self.workers.items():
            workers_conf[worker_name] = {}
            use_cuda = (worker_desc["type"] == "GPU")
            use_opencl = not(use_cuda) #
            workers_conf[worker_name]["use_cuda"] = use_cuda
            workers_conf[worker_name]["use_opencl"] = use_opencl
            self.workers[worker_name]["use_cuda"] = use_cuda
            self.workers[worker_name]["use_opencl"] = use_opencl
        # Broadcast
        self.client.datasets.update(workers_conf)


    def _create_tasks(self):
        cfg = self.process_config.nabu_config
        start_z = cfg["reconstruction"]["start_z"]
        end_z = cfg["reconstruction"]["end_z"]
        self._configure_chunk_size()
        tasks = []
        # method == "same"
        # TODO implement other methods
        chunk_size = self._common_chunk_size
        n_chunks = ceil((end_z - start_z)/chunk_size)
        for chunk_id in range(n_chunks):
            sub_region = (
                None, None,
                start_z + chunk_id * chunk_size,
                min(end_z, start_z + (chunk_id + 1) * chunk_size)
            )
            # ~ tasks.append((sub_region, chunk_size, chunk_id))
            tasks.append((sub_region, chunk_id))
        self._tasks = tasks


    def _spawn_workers_pipelines(self):
        actors = {}
        for worker_name, worker_conf in self.workers.items():
            fut = self.client.submit(
                WorkerProcess,
                self.process_config,
                (None, None, 0, worker_conf["chunk_size"]), # placeholder !
                chunk_size=worker_conf["chunk_size"],
                use_cuda=worker_conf["use_cuda"],
                use_opencl=worker_conf["use_opencl"],
                logger=None, # TODO
                extra_options={"clear_gpu_memory_when_possible": True},
                workers=[worker_conf["address"]],
                actor=True
            )
            actors[worker_name] = fut.result()
        self._actors = actors


    def get_workers_type(self, worker_type):
        """
        Return the address of a type of worker.

        Parameters
        -----------
        worker_type: str
            Type of worker. Can be "CPU" or "GPU".
        """
        return list(filter(lambda x: x["type"] == worker_type, self.workers.values()))


    # WIP !
    def reconstruct_volume(self):
        futures = []
        for task in self._tasks:
            f = self.client.submit(
                # ~ worker_process_chunk,
                actor_process_chunk,
                *task
            )
            futures.append(f)
        self.futures = futures



'''

    LOCAL
    -------
    memory needed
        gpu worker: 2.5*gpu mem
        cpu worker: avail_mem / n_cpu_workers

    limit:
        upper_tot = min(sys_avail_mem, user_requested_mem["memory_per_node"])
        upper_worker = upper_tot/n_workers


    DISTRIBUTED
    -----------
    one gpu worker comes with its cpu companion
    we assume there are 2 gpus per node, so we take sys_avail_mem/2 at most




'''



from ..reconstruction.sinogram import *
from ..utils import deprecation_warning

deprecation_warning(
    "nabu.preproc.sinogram was moved to nabu.reconstruction.sinogram.",
    func_name="preproc_sinogram"
)


import numpy as np

from .alignment import plt
from .cor import CenterOfRotation

class CameraFocus(CenterOfRotation):
    def find_distance(
        self,
        img_stack: np.ndarray,
        img_pos: np.array,
        metric="std",
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        peak_fit_radius=1,
        high_pass=None,
        low_pass=None,
    ):
        """Find the focal distance of the camera system.

        This routine computes the motor position that corresponds to having the
        scintillator on the focal plain of the camera system.

        Parameters
        ----------
        img_stack: numpy.ndarray
            A stack of images at different distances.
        img_pos: numpy.ndarray
            Position of the images along the translation axis
        metric: string, optional
            The property, whose maximize occurs at the focal position.
            Defaults to 'std' (standard deviation).
            All options are: 'std' | 'grad' | 'psd'
        roi_yxhw: (2, ) or (4, ) numpy.ndarray, tuple, or array, optional
            4 elements vector containing: vertical and horizontal coordinates
            of first pixel, plus height and width of the Region of Interest (RoI).
            Or a 2 elements vector containing: plus height and width of the
            centered Region of Interest (RoI).
            Default is None -> deactivated.
        median_filt_shape: (2, ) numpy.ndarray, tuple, or array, optional
            Shape of the median filter window. Default is None -> deactivated.
        padding_mode: str in numpy.pad's mode list, optional
            Padding mode, which determines the type of convolution. If None or
            'wrap' are passed, this resorts to the traditional circular convolution.
            If 'edge' or 'constant' are passed, it results in a linear convolution.
            Default is the circular convolution.
            All options are:
                None | 'constant' | 'edge' | 'linear_ramp' | 'maximum' | 'mean'
                | 'median' | 'minimum' | 'reflect' | 'symmetric' |'wrap'
        peak_fit_radius: int, optional
            Radius size around the max correlation pixel, for sub-pixel fitting.
            Minimum and default value is 1.
        low_pass: float or sequence of two floats
            Low-pass filter properties, as described in `nabu.misc.fourier_filters`.
        high_pass: float or sequence of two floats
            High-pass filter properties, as described in `nabu.misc.fourier_filters`.

        Returns
        -------
        focus_pos: float
            Estimated position of the focal plane of the camera system.
        focus_ind: float
            Image index of the estimated position of the focal plane of the camera system (starting from 1!).

        Examples
        --------
        Given the focal stack associated to multiple positions of the camera
        focus motor called `img_stack`, and the associated positions `img_pos`,
        the following code computes the highest focus position:

        >>> focus_calc = alignment.CameraFocus()
        ... focus_pos, focus_ind = focus_calc.find_distance(img_stack, img_pos)

        where `focus_pos` is the corresponding motor position, and `focus_ind`
        is the associated image position (starting from 1).
        """
        self._check_img_stack_size(img_stack, img_pos)

        if peak_fit_radius < 1:
            self.logger.warning("Parameter peak_fit_radius should be at least 1, given: %d instead." % peak_fit_radius)
            peak_fit_radius = 1

        num_imgs = img_stack.shape[0]
        img_shape = img_stack.shape[-2:]
        roi_yxhw = self._determine_roi(img_shape, roi_yxhw)

        img_stack = self._prepare_image(
            img_stack, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape, low_pass=low_pass, high_pass=high_pass,
        )

        img_stds = np.std(img_stack, axis=(-2, -1)) / np.mean(img_stack, axis=(-2, -1))

        # assuming images are equispaced
        focus_step = (img_pos[-1] - img_pos[0]) / (num_imgs - 1)

        img_inds = np.arange(num_imgs)
        (f_vals, f_pos) = self.extract_peak_regions_1d(img_stds, peak_radius=peak_fit_radius, cc_coords=img_inds)
        focus_ind, img_std_max = self.refine_max_position_1d(f_vals, return_vertex_val=True)
        focus_ind += f_pos[1, :]

        focus_pos = img_pos[0] + focus_step * focus_ind
        focus_ind += 1

        if self.verbose:
            self.logger.info(
                "Fitted focus motor position:", focus_pos, "and corresponding image position:", focus_ind,
            )
            f, ax = plt.subplots(1, 1)
            self._add_plot_window(f, ax=ax)
            ax.stem(img_pos, img_stds)
            ax.stem(focus_pos, img_std_max, linefmt="C1-", markerfmt="C1o")
            ax.set_title("Images std")
            plt.show(block=False)

        return focus_pos, focus_ind

    def _check_img_block_size(self, img_shape, regions_number, suggest_new_shape=True):
        img_shape = np.array(img_shape)
        new_shape = img_shape
        if not len(img_shape) == 2:
            raise ValueError(
                "Images need to be square 2-dimensional and with shape multiple of the number of assigned regions.\n"
                " Image shape: %s, regions number: %d" % (img_shape, regions_number)
            )
        if not (img_shape[0] == img_shape[1] and np.all((np.array(img_shape) % regions_number) == 0)):
            new_shape = (img_shape // regions_number) * regions_number
            new_shape = np.fmin(new_shape, new_shape.min())
            message = (
                "Images need to be square 2-dimensional and with shape multiple of the number of assigned regions.\n"
                " Image shape: %s, regions number: %d. Cropping to image shape: %s" % (img_shape, regions_number, new_shape)
            )
            if suggest_new_shape:
                self.logger.info(message)
            else:
                raise ValueError(message)
        return new_shape

    @staticmethod
    def _fit_plane(f_vals):
        f_vals_half_shape = (np.array(f_vals.shape) - 1) / 2

        fy = np.linspace(-f_vals_half_shape[-2], f_vals_half_shape[-2], f_vals.shape[-2])
        fx = np.linspace(-f_vals_half_shape[-1], f_vals_half_shape[-1], f_vals.shape[-1])

        fy, fx = np.meshgrid(fy, fx, indexing="ij")
        coords = np.array([np.ones(f_vals.size), fy.flatten(), fx.flatten()])

        return np.linalg.lstsq(coords.T, f_vals.flatten(), rcond=None)[0], fy, fx

    def find_scintillator_tilt(
        self,
        img_stack: np.ndarray,
        img_pos: np.array,
        regions_number=4,
        metric="std",
        roi_yxhw=None,
        median_filt_shape=None,
        padding_mode=None,
        peak_fit_radius=1,
        high_pass=None,
        low_pass=None,
    ):
        """Finds the scintillator tilt and focal distance of the camera system.

        This routine computes the mounting tilt of the scintillator and the
        motor position that corresponds to having the scintillator on the focal
        plain of the camera system.

        The input is supposed to be a stack of square images, whose sizes are
        multiples of the `regions_number` parameter. If images with a different
        size are passed, this function will crop the images. This also generates
        a warning. To suppress the warning, it is suggested to specify a ROI
        that satisfies those criteria (see examples).

        The computed tilts `tilt_vh` are in unit-length per pixel-size. To
        obtain the tilts it is necessary to divide by the pixel-size:

        >>> tilt_vh_deg = np.rad2deg(np.arctan(tilt_vh / pixel_size))

        The correction to be applied is:

        >>> tilt_corr_vh_deg = - np.rad2deg(np.arctan(tilt_vh / pixel_size))

        The legacy octave macros computed the approximation of these values in radians:

        >>> tilt_corr_vh_rad = - tilt_vh / pixel_size

        Note that `pixel_size` should be in the same unit scale as `img_pos`.

        Parameters
        ----------
        img_stack: numpy.ndarray
            A stack of images at different distances.
        img_pos: numpy.ndarray
            Position of the images along the translation axis
        regions_number: int, optional
            The number of regions to subdivide the image into, along each direction.
            Defaults to 4.
        metric: string, optional
            The property, whose maximize occurs at the focal position.
            Defaults to 'std' (standard deviation).
            All options are: 'std' | 'grad' | 'psd'
        roi_yxhw: (2, ) or (4, ) numpy.ndarray, tuple, or array, optional
            4 elements vector containing: vertical and horizontal coordinates
            of first pixel, plus height and width of the Region of Interest (RoI).
            Or a 2 elements vector containing: plus height and width of the
            centered Region of Interest (RoI).
            Default is None -> auto-suggest correct size.
        median_filt_shape: (2, ) numpy.ndarray, tuple, or array, optional
            Shape of the median filter window. Default is None -> deactivated.
        padding_mode: str in numpy.pad's mode list, optional
            Padding mode, which determines the type of convolution. If None or
            'wrap' are passed, this resorts to the traditional circular convolution.
            If 'edge' or 'constant' are passed, it results in a linear convolution.
            Default is the circular convolution.
            All options are:
                None | 'constant' | 'edge' | 'linear_ramp' | 'maximum' | 'mean'
                | 'median' | 'minimum' | 'reflect' | 'symmetric' |'wrap'
        peak_fit_radius: int, optional
            Radius size around the max correlation pixel, for sub-pixel fitting.
            Minimum and default value is 1.
        low_pass: float or sequence of two floats
            Low-pass filter properties, as described in `nabu.misc.fourier_filters`.
        high_pass: float or sequence of two floats
            High-pass filter properties, as described in `nabu.misc.fourier_filters`.

        Returns
        -------
        focus_pos: float
            Estimated position of the focal plane of the camera system.
        focus_ind: float
            Image index of the estimated position of the focal plane of the
            camera system (starting from 1!).
        tilts_vh: tuple(float, float)
            Estimated scintillator tilts in the vertical and horizontal
            direction respectively per unit-length per pixel-size.

        Examples
        --------
        Given the focal stack associated to multiple positions of the camera
        focus motor called `img_stack`, and the associated positions `img_pos`,
        the following code computes the highest focus position:

        >>> focus_calc = alignment.CameraFocus()
        ... focus_pos, focus_ind, tilts_vh = focus_calc.find_scintillator_tilt(img_stack, img_pos)
        ... tilt_corr_vh_deg = - np.rad2deg(np.arctan(tilt_vh / pixel_size))

        or to keep compatibility with the old octave macros:

        >>> tilt_corr_vh_rad = - tilt_vh / pixel_size

        For non square images, or images with sizes that are not multiples of
        the `regions_number` parameter, and no ROI is being passed, this function
        will try to crop the image stack to the correct size.
        If you want to remove the warning message, it is suggested to set a ROI
        like the following:

        >>> regions_number = 4
        ... img_roi = (np.array(img_stack.shape[1:]) // regions_number) * regions_number
        ... img_roi = np.fmin(img_roi, img_roi.min())
        ... focus_calc = alignment.CameraFocus()
        ... focus_pos, focus_ind, tilts_vh = focus_calc.find_scintillator_tilt(
        ...     img_stack, img_pos, roi_yxhw=img_roi, regions_number=regions_number)
        """
        self._check_img_stack_size(img_stack, img_pos)

        if peak_fit_radius < 1:
            self.logger.warning("Parameter peak_fit_radius should be at least 1, given: %d instead." % peak_fit_radius)
            peak_fit_radius = 1

        num_imgs = img_stack.shape[0]
        img_shape = img_stack.shape[-2:]
        if roi_yxhw is None:
            # If no roi is being passed, we try to crop the images to the
            # correct size, if needed
            roi_yxhw = self._check_img_block_size(img_shape, regions_number, suggest_new_shape=True)
            roi_yxhw = self._determine_roi(img_shape, roi_yxhw)
        else:
            # If a roi is being passed, and the images don't have the correct
            # shape, we raise an error
            roi_yxhw = self._determine_roi(img_shape, roi_yxhw)
            self._check_img_block_size(roi_yxhw[2:], regions_number, suggest_new_shape=False)

        img_stack = self._prepare_image(
            img_stack, roi_yxhw=roi_yxhw, median_filt_shape=median_filt_shape, low_pass=low_pass, high_pass=high_pass,
        )
        img_shape = img_stack.shape[-2:]

        block_size = np.array(img_shape) / regions_number
        block_stack_size = np.array([num_imgs, regions_number, block_size[-2], regions_number, block_size[-1]], dtype=np.intp,)
        img_stack = np.reshape(img_stack, block_stack_size)

        img_stds = np.std(img_stack, axis=(-3, -1)) / np.mean(img_stack, axis=(-3, -1))
        img_stds = np.reshape(img_stds, [num_imgs, -1]).transpose()

        # assuming images are equispaced
        focus_step = (img_pos[-1] - img_pos[0]) / (num_imgs - 1)

        img_inds = np.arange(num_imgs)
        (f_vals, f_pos) = self.extract_peak_regions_1d(img_stds, peak_radius=peak_fit_radius, cc_coords=img_inds)
        focus_inds = self.refine_max_position_1d(f_vals)
        focus_inds += f_pos[1, :]

        focus_poss = img_pos[0] + focus_step * focus_inds

        # Fitting the plane
        focus_poss = np.reshape(focus_poss, [regions_number, regions_number])
        coeffs, fy, fx = self._fit_plane(focus_poss)
        focus_pos, tg_v, tg_h = coeffs

        # The angular coefficient along x is the tilt around the y axis and vice-versa
        tilts_vh = np.array([tg_h, tg_v]) / block_size

        focus_ind = np.mean(focus_inds) + 1

        if self.verbose:
            self.logger.info(
                "Fitted focus motor position:", focus_pos, "and corresponding image position:", focus_ind,
            )
            self.logger.info("Fitted tilts (to be divided by pixel size, and converted to deg): (v, h) %s" % tilts_vh)
            fig = plt.figure()
            ax = fig.add_subplot(111, projection="3d")
            self._add_plot_window(fig, ax=ax)
            ax.plot_wireframe(fx, fy, focus_poss)
            regions_half_shape = (regions_number - 1) / 2
            base_points = np.linspace(-regions_half_shape, regions_half_shape, regions_number)
            ax.plot(
                np.zeros((regions_number,)), base_points, np.polyval([tg_v, focus_pos], base_points), "C2",
            )
            ax.plot(
                base_points, np.zeros((regions_number,)), np.polyval([tg_h, focus_pos], base_points), "C2",
            )
            ax.scatter(0, 0, focus_pos, marker="o", c="C1")
            ax.set_title("Images std")
            plt.show(block=False)

        return focus_pos, focus_ind, tilts_vh



from os import path
from .utils import parse_params_values
from .cli_configs import BootstrapConfig
from ..pipeline.config import generate_nabu_configfile
from ..pipeline.fullfield.nabu_config import nabu_config as default_fullfield_config

def parse_sections(sections):
    sections = sections.lower()
    if sections == "all":
        return None
    sections = sections.replace(" ", "").split(",")
    return sections


def bootstrap():
    args = parse_params_values(
        BootstrapConfig,
        parser_description="Initialize a nabu configuration file"
    )

    do_bootstrap = bool(args["bootstrap"])
    do_convert = (args["convert"] != "")
    no_comments = bool(args["nocomments"])

    if do_convert: # do_convert
        print("The --convert option is deprecated. It will be removed in a future version")
        exit(1)
    if do_bootstrap:
        print("The --bootstrap option is now the default behavior of the nabu-config command. This option is therefore not needed anymore.")

    if path.isfile(args["output"]):
        rep = input("File %s already exists. Overwrite ? [y/N]" % args["output"])
        if rep.lower() != "y":
            print("Stopping")
            exit(0)

    prefilled_values = {}
    if args["dataset"] != "":
        prefilled_values["dataset"] = {}
        user_dataset = args["dataset"]
        if not path.isabs(user_dataset):
            user_dataset = path.abspath(user_dataset)
            print("Warning: using absolute dataset path %s" % user_dataset)
        if not path.exists(user_dataset):
            print("Error: cannot find the file or directory %s" % user_dataset)
            exit(1)
        prefilled_values["dataset"]["location"] = user_dataset

    generate_nabu_configfile(
        args["output"],
        default_fullfield_config,
        comments=not(no_comments),
        options_level=args["level"],
        prefilled_values=prefilled_values,
    )



import numpy as np
from .double_flatfield import DoubleFlatField
from ..utils import check_shape
from ..cuda.utils import get_cuda_context, __has_pycuda__
from ..cuda.processing import CudaProcessing
from ..misc.unsharp_cuda import CudaUnsharpMask

if __has_pycuda__:
    import pycuda.gpuarray as garray
    import pycuda.cumath as cumath

class CudaDoubleFlatField(DoubleFlatField):

    def __init__(
        self,
        shape,
        result_url=None,
        sub_region=None,
        input_is_mlog=True,
        output_is_mlog=False,
        average_is_on_log=False,
        sigma_filter=None,
        filter_mode="reflect",
        cuda_options=None,
    ):
        """
        Init double flat field with Cuda backend.
        """
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        super().__init__(
            shape,
            result_url=result_url,
            sub_region=sub_region,
            input_is_mlog=input_is_mlog,
            output_is_mlog=output_is_mlog,
            average_is_on_log=average_is_on_log,
            sigma_filter=sigma_filter,
            filter_mode=filter_mode
        )
        self._init_gaussian_filter()


    def _init_gaussian_filter(self):
        if self.sigma_filter is None:
            return
        self._unsharp_mask = CudaUnsharpMask(
            self.shape, self.sigma_filter, -1.,
            mode=self.filter_mode, method="log"
        )


    @staticmethod
    def _proc_copy(x, o):
        o[:] = x[:]
        return o

    @staticmethod
    def _proc_expm(x, o):
        o[:] = x[:]
        o[:] *= -1
        cumath.exp(o, out=o)
        return o

    @staticmethod
    def _proc_mlog(x, o, min_clip=None):
        if min_clip is not None:
            garray.maximum(x, min_clip, output=o)
            cumath.log(o, out=o)
        else:
            cumath.log(x, out=o)
        o *= -1
        return o


    def _init_processing(self, input_is_mlog, output_is_mlog, average_is_on_log, sigma_filter, filter_mode):
        self.input_is_mlog = input_is_mlog
        self.output_is_mlog = output_is_mlog
        self.average_is_on_log = average_is_on_log
        self.sigma_filter = sigma_filter
        if self.sigma_filter is not None and abs(float(self.sigma_filter)) < 1e-4:
            self.sigma_filter = None
        self.filter_mode = filter_mode
        # proc = lambda x,o: np.copyto(o, x)
        proc = self._proc_copy
        if self.input_is_mlog:
            if not self.average_is_on_log:
                # proc = lambda x,o: np.exp(-x, out=o)
                proc = self._proc_expm
        else:
            if self.average_is_on_log:
                # proc = lambda x,o: -np.log(x, out=o)
                proc = self._proc_mlog

        # postproc = lambda x: x
        postproc = self._proc_copy
        if self.output_is_mlog:
            if not self.average_is_on_log:
                # postproc = lambda x: -np.log(x)
                postproc = self._proc_mlog
        else:
            if self.average_is_on_log:
                # postproc = lambda x: np.exp(-x)
                postproc = self._proc_expm

        self.proc = proc
        self.postproc = postproc


    def compute_double_flatfield(self, radios, recompute=False):
        """
        Read the radios and generate the "double flat field" by averaging
        and possibly other processing.

        Parameters
        ----------
        radios: array
            Input radios chunk.
        recompute: bool, optional
            Whether to recompute the double flatfield if already computed.
        """
        if not(isinstance(radios, garray.GPUArray)):
            raise ValueError("Expected pycuda.gpuarray.GPUArray for radios")
        if self._computed:
            return self.doubleflatfield
        acc = garray.zeros(radios[0].shape, "f")
        tmpdat = garray.zeros(radios[0].shape, "f")
        for i in range(radios.shape[0]):
            self.proc(radios[i], tmpdat)
            acc += tmpdat

        acc /= radios.shape[0]

        if self.sigma_filter is not None:
            # acc = acc - gaussian_filter(acc, self.sigma_filter, mode=self.filter_mode)
            self._unsharp_mask.unsharp(acc, tmpdat)
            acc[:] = tmpdat[:]
        self.postproc(acc, tmpdat)
        self.doubleflatfield = tmpdat
        # Handle small values to avoid issues when dividing
        # self.doubleflatfield[np.abs(self.doubleflatfield) < self._small] = 1.
        cumath.fabs(self.doubleflatfield, out=acc)
        acc -= self._small # acc = abs(doubleflatfield) - _small
        garray.if_positive(
            acc,
            self.doubleflatfield,
            garray.zeros_like(acc) + self._small,
            out=self.doubleflatfield
        )

        if self.writer is not None:
            self.writer.write(self.doubleflatfield.get(), "double_flatfield")
        self._computed = True
        return self.doubleflatfield


    def get_double_flatfield(self, radios=None, compute=False):
        """
        Get the double flat field or a subregion of it.

        Parameters
        ----------
        radios: array, optional
            Input radios chunk
        compute: bool, optional
            Whether to compute the double flatfield anyway even if a dump file
            exists.
        """
        if self.reader is None:
            if radios is None:
                raise ValueError(
                    "result_url was not provided. Please provide 'radios' to this function"
                )
            return self.compute_double_flatfield(radios)

        if radios is not None and compute:
            res = self.compute_double_flatfield(radios)
        else:
            res = self._load_dff_dump()
            res = garray.to_gpu(res)
            self._computed = True
        return res


    def apply_double_flatfield(self, radios):
        """
        Apply the "double flatfield" filter on a chunk of radios.
        The processing is done in-place !
        """
        check_shape(radios.shape, self.radios_shape, "radios")
        dff = self.get_double_flatfield(radios=radios)
        for i in range(self.n_angles):
            radios[i] /= dff
        return radios



from argparse import ArgumentParser

def parse_params_values(Params, parser_description=None, program_version=None):
    parser = ArgumentParser(description=parser_description)
    for param_name, vals in Params.items():
        optional = not(vals.pop("mandatory", False))
        if optional:
            param_name = "--" + param_name
        parser.add_argument(param_name, **vals)
    if program_version is not None:
        parser.add_argument('--version', '-V', action='version', version=program_version)
    args = parser.parse_args()
    args_dict = args.__dict__
    return args_dict



import numpy as np
from .utils import get_cuda_context, __has_pycuda__

if __has_pycuda__:
    import pycuda.driver as cuda
    import pycuda.gpuarray as garray
    dev_attrs = cuda.device_attribute

# NB: we must detach from a context before creating another context
class CudaProcessing:
    def __init__(self, device_id=None, ctx=None, stream=None, cleanup_at_exit=True):
        """
        Initialie a CudaProcessing instance.

        CudaProcessing is a base class for all CUDA-based processings.
        This class provides utilities for context/device management, and
        arrays allocation.

        Parameters
        ----------
        device_id: int, optional
            ID of the cuda device to use (those of the `nvidia-smi` command).
            Ignored if ctx is not None.
        ctx: pycuda.driver.Context, optional
            Existing context to use. If provided, do not create a new context.
        stream: pycudacuda.driver.Stream, optional
            Cuda stream. If not provided, will use the default stream
        cleanup_at_exit: bool, optional
            Whether to clean-up the context at exit.
            Ignored if ctx is not None.
        """
        if ctx is None:
            self.ctx = get_cuda_context(
                device_id=device_id,
                cleanup_at_exit=cleanup_at_exit
            )
        else:
            self.ctx = ctx
        self.stream = stream
        self.device = self.ctx.get_device()
        self.device_name = self.device.name()
        self.device_id = self.device.get_attribute(dev_attrs.MULTI_GPU_BOARD_GROUP_ID)
        self._allocated = {}


    def push_context(self):
        self.ctx.push()
        return self.ctx


    def pop_context(self):
        self.ctx.pop()


    def init_arrays_to_none(self, arrays_names):
        """
        Initialize arrays to None. After calling this method, the current instance will
        have self.array_name = None, and self._old_array_name = None.

        Parameters
        ----------
        arrays_names: list of str
            List of arrays names.
        """
        for array_name in arrays_names:
            setattr(self, array_name, None)
            setattr(self, "_old_" + array_name, None)
            self._allocated[array_name] = False


    def recover_arrays_references(self, arrays_names):
        """
        Performs self._array_name = self._old_array_name,
        for each array_name in arrays_names.

        Parameters
        ----------
        arrays_names: list of str
            List of array names
        """
        for array_name in arrays_names:
            old_arr = getattr(self, "_old_" + array_name)
            if old_arr is not None:
                setattr(self, array_name, old_arr)


    def allocate_array(self, array_name, shape, dtype=np.float32):
        """
        Allocate a GPU array on the current context/stream/device,
        and set 'self.array_name' to this array.

        Parameters
        ----------
        array_name: str
            Name of the array (for book-keeping)
        shape: tuple of int
            Array shape
        dtype: numpy.dtype, optional
            Data type. Default is float32.
        """
        if not self._allocated.get(array_name, False):
            new_gpuarr = garray.zeros(shape, dtype=dtype)
            setattr(self, array_name, new_gpuarr)
            self._allocated[array_name] = True
        return getattr(self, array_name)


    def set_array(self, array_name, array_ref, shape, dtype=np.float32):
        """
        Set the content of a device array.

        Parameters
        ----------
        array_name: str
            Array name. This method will look for self.array_name.
        array_ref: array (numpy or GPU array)
            Array containing the data to copy to 'array_name'.
        shape: tuple of int
            Array shape
        dtype: numpy.dtype, optional
            Data type. Default is float32.
        """
        if isinstance(array_ref, garray.GPUArray):
            current_arr = getattr(self, array_name)
            setattr(self, "_old_" + array_name, current_arr)
            setattr(self, array_name, array_ref)
        elif isinstance(array_ref, np.ndarray):
            self._allocate_array(array_name, shape, dtype=dtype)
            getattr(self, array_name).set(array_ref)
        else:
            raise ValueError("Expected numpy array or pycuda array")

    # COMPAT.
    _init_arrays_to_none = init_arrays_to_none
    _recover_arrays_references = recover_arrays_references
    _allocate_array = allocate_array
    _set_array = set_array



#
# Default configuration for CLI tools
#

# Default configuration for "bootstrap" command
BootstrapConfig = {
    "bootstrap": {
        "help": "DEPRECATED, this is the default behavior. Bootstrap a configuration file from scratch.",
        "action": "store_const",
        "const": 1,
    },
    "convert": {
        "help": "UNSUPPORTED. This option has no effect and will disappear. Convert a PyHST configuration file to a nabu configuration file.",
        "default": "",
    },
    "output": {
        "help": "Output filename",
        "default": "nabu.conf",
    },
    "nocomments": {
        "help": "Remove the comments in the configuration file (default: False)",
        "action": "store_const",
        "const": 1,
    },
    "level": {
        "help": "Level of options to embed in the configuration file. Can be 'required', 'optional', 'advanced'.",
        "default": "optional",
    },
    "dataset": {
        "help": "Pre-fill the configuration file with the dataset path.",
        "default": "",
    }
}

# Default configuration for "zsplit" command
ZSplitConfig = {
    "input_file": {
        "help": "Input HDF5-Nexus file",
        "mandatory": True,
    },
    "output_directory": {
        "help": "Output directory to write split files.",
        "mandatory": True,
    },
    "loglevel": {
        "help": "Logging level. Can be 'debug', 'info', 'warning', 'error'. Default is 'info'.",
        "default": "info",
    },
    "entry": {
        "help": "HDF5 entry to take in the input file. By default, the first entry is taken.",
        "default": "",
    },
    "n_stages": {
        "help": "Number of expected stages (i.e different 'Z' values). By default it is inferred from the dataset.",
        "default": -1,
        "type": int,
    },
    "use_virtual_dataset": {
        "help": "Whether to use virtual datasets for output file. Not using a virtual dataset duplicates data and thus results in big files ! However virtual datasets currently have performance issues. Default is False",
        "default": 0,
        "type": int,
    },
}

# Default configuration for "histogram" command
HistogramConfig = {
    "h5_file": {
        "help": "HDF5 file(s). It can be one or several paths to HDF5 files. You can specify entry for each file with /path/to/file.h5?entry0000",
        "mandatory": True,
        "nargs": "+",
    },
    "output_file": {
        "help": "Output file (HDF5)",
        "mandatory": True,
    },
    "bins": {
        "help": "Number of bins for histogram if they have to be computed. Default is one million.",
        "default": 1000000,
        "type": int,
    },
    "chunk_size_slices": {
        "help": "If histogram are computed, specify the maximum subvolume size (in number of slices) for computing histogram.",
        "default": 100,
        "type": int,
    },
    "chunk_size_GB": {
        "help": "If histogram are computed, specify the maximum subvolume size (in GibaBytes) for computing histogram.",
        "default": -1,
        "type": float,
    },
    "loglevel": {
        "help": "Logging level. Can be 'debug', 'info', 'warning', 'error'. Default is 'info'.",
        "default": "info",
    },
}


# Default configuration for "reconstruct" command
ReconstructConfig = {
    "input_file": {
        "help": "Nabu input file",
        "default": "",
        "mandatory": True,
    },
    "logfile": {
        "help": "Log file. Default is dataset_prefix_nabu.log",
        "default": "",
    },
    "log_file": {
        "help": "Same as logfile. Deprecated, use --logfile instead.",
        "default": "",
    },
    "slice": {
        "help": "Slice(s) indice(s) to reconstruct, in the format z1-z2. Default (empty) is the whole volume. This overwrites the configuration file start_z and end_z. You can also use --slice first, --slice last, --slice middle, and --slice all",
        "default": "",
    },
    # ~ "compute": {
        # ~ "help": "Computation distribution method. Can be 'local' or 'slurm'",
        # ~ "default": "local",
    # ~ },
    # ~ "nodes": {
        # ~ "help": "Number of computing nodes to use. Ignored if --compute is set to 'local'.",
        # ~ "default": 1,
        # ~ "type": int,
    # ~ },
    "energy": {
        "help": "Beam energy in keV. DEPRECATED, was used to patch missing fields in BCU HDF5 file.",
        "default": -1,
        "type": float,
    },
    "gpu_mem_fraction": {
        "help": "Which fraction of GPU memory to use. Default is 0.9.",
        "default": 0.9,
        "type": float,
    },
    "cpu_mem_fraction": {
        "help": "Which fraction of memory to use. Default is 0.9.",
        "default": 0.9,
        "type": float,
    },
    "max_chunk_size": {
        "help": "Maximum chunk size to use.",
        "default": -1,
        "type": int,
    },
    "use_phase_margin": {
        "help": "DEPRECATED. Use --phase_margin instead. Whether to use a margin when performing phase retrieval. Defautl is True.",
        "default": True,
        "type": bool,
    },
    "phase_margin": {
        "help": "Specify an explicit phase margin to use when performing phase retrieval.",
        "default": -1,
        "type": int,
    },
    "force_use_grouped_pipeline": {
        "help": "Force nabu to use the 'grouped' reconstruction pipeline - slower but should work for all big datasets.",
        "default": 0,
        "type": int,
    },
}

GenerateInfoConfig = {
    "hist_file": {
        "help": "HDF5 file containing the histogram, either the reconstruction file or a dedicated histogram file.",
        "default": "",
    },
    "hist_entry": {
        "help": "Histogram HDF5 entry. Defaults to the first available entry.",
        "default": "",
    },
    "output": {
        "help": "Output file name",
        "default": "",
        "mandatory": True,
    },
    "bliss_file": {
        "help": "HDF5 master file produced by BLISS",
        "default": "",
    },
    "bliss_entry": {
        "help": "Entry in the HDF5 master file produced by BLISS. By default, take the first entry.",
        "default": "",
    },
    "info_file": {
        "help": "Path to the .info file, in the case of a EDF dataset",
        "default": "",
    },
    "edf_proj": {
        "help": "Path to a projection, in the case of a EDF dataset",
        "default": "",
    },

}

RotateRadiosConfig = {
    "dataset": {
        "help": "Path to the dataset. Only HDF5 format is supported for now.",
        "default": "",
        "mandatory": True,
    },
    "entry": {
        "help": "HDF5 entry. By default, the first entry is taken.",
        "default": "",
    },
    "angle": {
        "help": "Rotation angle in degrees",
        "default": 0.,
        "mandatory": True,
        "type": float,
    },
    "center": {
        "help": "Rotation center, in the form (x, y) where x (resp. y) is the horizontal (resp. vertical) dimension, i.e along the columns (resp. lines). Default is (Nx/2 - 0.5, Ny/2 - 0.5).",
        "default": "",
    },
    "output": {
        "help": "Path to the output file. Only HDF5 output is supported. In the case of HDF5 input, the output file will have the same structure.",
        "default": "",
        "mandatory": True,
    },
    "loglevel": {
        "help": "Logging level. Can be 'debug', 'info', 'warning', 'error'. Default is 'info'.",
        "default": "info",
    },
    "batchsize": {
        "help": "Size of the batch of images to process. Default is 100",
        "default": 100,
        "type": int,
    },
    "use_cuda": {
        "help": "Whether to use Cuda if available",
        "default": "1",
    },
    "use_multiprocessing": {
        "help": "Whether to use multiprocessing if available",
        "default": "1",
    },
}

DFFConfig = {
    "dataset": {
        "help": "Path to the dataset.",
        "default": "",
        "mandatory": True,
    },
    "entry": {
        "help": "HDF5 entry (for HDF5 datasets). By default, the first entry is taken.",
        "default": "",
    },
    "flatfield": {
        "help": "Whether to perform flat-field normalization. Default is True.",
        "default": "1",
        "type": int,
    },
    "sigma": {
        "default": 0.,
        "help": "Enable high-pass filtering on double flatfield with this value of 'sigma'",
        "type": float,
    },
    "output": {
        "help": "Path to the output file (HDF5).",
        "default": "",
        "mandatory": True,
    },
    "loglevel": {
        "help": "Logging level. Can be 'debug', 'info', 'warning', 'error'. Default is 'info'.",
        "default": "info",
    },
    "chunk_size": {
        "help": "Maximum number of lines to read in each projection in a single pass. Default is 100",
        "default": 100,
        "type": int,
    },
}



import numpy as np
from math import sqrt, pi
from ..utils import updiv, get_cuda_srcfile, _sizeof, check_supported
from .phase import PaganinPhaseRetrieval
import pycuda.driver as cuda
from pycuda import gpuarray as garray
from ..cuda.processing import CudaProcessing
from ..cuda.kernel import CudaKernel


class CudaPaganinPhaseRetrieval(PaganinPhaseRetrieval):

    supported_paddings = ["zeros", "constant", "edge"]

    def __init__(
        self,
        shape,
        distance=0.5,
        energy=20,
        delta_beta=250.,
        pixel_size=1e-6,
        padding="edge",
        margin=None,
        cuda_options=None,
        fftw_num_threads=None,
    ):
        """
        Please refer to the documentation of
        nabu.preproc.phase.PaganinPhaseRetrieval
        """
        padding = self._check_padding(padding)
        self.cuda_processing = CudaProcessing(**(cuda_options or {}))
        super().__init__(
            shape,
            distance=distance,
            energy=energy,
            delta_beta=delta_beta,
            pixel_size=pixel_size,
            padding=padding,
            margin=margin,
            use_R2C=True,
            fftw_num_threads=None,
        )
        self._init_gpu_arrays()
        self._init_fft()
        self._init_padding_kernel()
        self._init_mult_kernel()


    def _check_padding(self, padding):
        check_supported(padding, self.supported_paddings, "padding")
        if padding == "zeros":
            padding = "constant"
        return padding


    def _init_gpu_arrays(self):
        self.d_paganin_filter = garray.to_gpu(
            np.ascontiguousarray(self.paganin_filter, dtype=np.float32)
        )


    def _init_fft(self):
        # Import has to be done here, otherwise scikit-cuda creates a cuda/cublas context at import
        from silx.math.fft.cufft import CUFFT
        #
        self.cufft = CUFFT(template=self.data_padded.astype("f"))
        self.d_radio_padded = self.cufft.data_in
        self.d_radio_f = self.cufft.data_out


    def _init_padding_kernel(self):
        kern_signature = {
            "constant": "Piiiiiiiiffff",
            "edge": "Piiiiiiii"
        }
        self.padding_kernel = CudaKernel(
            "padding_%s" % self.padding,
            filename=get_cuda_srcfile("padding.cu"),
            signature=kern_signature[self.padding],
        )
        Ny, Nx = self.shape
        Nyp, Nxp = self.shape_padded
        self.padding_kernel_args = [
            self.d_radio_padded,
            Nx, Ny,
            Nxp, Nyp,
            self.pad_left_len, self.pad_right_len,
            self.pad_top_len, self.pad_bottom_len,
        ]
        # TODO configurable constant values
        if self.padding == "constant":
            self.padding_kernel_args.extend([
                0, 0, 0, 0
            ])


    def _init_mult_kernel(self):
        self.cpxmult_kernel = CudaKernel(
            "inplace_complexreal_mul_2Dby2D",
            filename=get_cuda_srcfile("ElementOp.cu"),
            signature="PPii",
        )
        self.cpxmult_kernel_args = [
            self.d_radio_f,
            self.d_paganin_filter,
            self.shape_padded[1]//2+1,
            self.shape_padded[0]
        ]


    def set_input(self, data):
        assert data.shape == self.shape
        assert data.dtype == np.float32
        # Rectangular memcopy
        # TODO profile, and if needed include this copy in the padding kernel
        if isinstance(data, np.ndarray) or isinstance(data, garray.GPUArray):
            self.d_radio_padded[:self.shape[0], :self.shape[1]] = data[:, :]
        elif isinstance(data, cuda.DeviceAllocation):
            # TODO manual memcpy2D
            raise NotImplementedError("pycuda buffers are not supported yet")
        else:
            raise ValueError("Expected either numpy array, pycuda array or pycuda buffer")


    def get_output(self, output):
        s0, s1 = self.shape_inner
        ((U, _), (L, _)) = self.margin
        if output is None:
            # copy D2H
            return  self.d_radio_padded[U:U + s0, L:L + s1].get()
        assert output.shape == self.shape_inner
        assert output.dtype == np.float32
        output[:, :] = self.d_radio_padded[U:U + s0, L:L + s1]
        return output


    def apply_filter(self, radio, output=None):
        self.set_input(radio)

        self.padding_kernel(*self.padding_kernel_args)
        self.cufft.fft(self.d_radio_padded, output=self.d_radio_f)
        self.cpxmult_kernel(*self.cpxmult_kernel_args)
        self.cufft.ifft(self.d_radio_f, output=self.d_radio_padded)

        return self.get_output(output)


    __call__ = apply_filter



#!/usr/bin/env python
# coding: utf-8

from setuptools import setup, find_packages
import os
from nabu import version, __nabu_modules__

def setup_package():
    doc_requires = [
        'sphinx',
        'cloud_sptheme',
        'myst-parser',
        'nbsphinx',
    ]
    setup(
        name='nabu',
        author='Pierre Paleo',
        version=version,
        author_email = "pierre.paleo@esrf.fr",
        maintainer = "Pierre Paleo",
        maintainer_email = "pierre.paleo@esrf.fr",

        packages=find_packages(),
        package_data = {
            'nabu.cuda': [
                'src/*.cu',
                'src/*.h',
            ],
            'nabu.resources': [
                'templates/*.ini',
            ],
        },
        include_package_data=True,

        install_requires = [
            'psutil',
            'pytest',
            'numpy > 1.9.0',
            'scipy',
            'silx >= 0.15.0',
            'distributed',
            'dask_jobqueue',
            'tomoscan >= 0.7.3',
            'h5py',
            'tifffile',
        ],
        description = "Nabu - Tomography software",

        entry_points = {
            'console_scripts': [
                "nabu-test=nabu.tests:nabu_test",
                "nabu-config=nabu.app.bootstrap:bootstrap",
                "nabu-zsplit=nabu.app.nx_z_splitter:zsplit",
                "nabu-histogram=nabu.app.histogram:histogram_cli",
                "nabu-rotate=nabu.app.rotate:rotate_cli",
                "nabu-double-flatfield=nabu.app.double_flatfield:dff_cli",
                "nabu-generate-info=nabu.app.generate_header:generate_merged_info_file",
                "nabu-validator=nabu.app.validator:main",
                "nabu=nabu.app.reconstruct:main",
            ],
        },

        zip_safe=True
    )


if __name__ == "__main__":
    setup_package()



"""
script embedding all function needed to read flats and darks.

For information calculation method is stored in the results/configuration
dictionary
"""


"""
IMPORTANT: this script is used as long as flat-fielding with "raw" flats/daks
is not implemented in nabu.
For now we load results from tomwer.
"""

import typing
import h5py
from silx.io.url import DataUrl
from tomoscan.io import HDF5File
from silx.io.utils import h5py_read_dataset
import logging

MAX_DEPTH = 2

logger = logging.getLogger(__name__)

def get_process_entries(root_node: h5py.Group, depth: int) -> tuple:
    """
    return the list of 'Nxtomo' entries at the root level

    :param str file_path:
    :return: list of valid Nxtomo node (ordered alphabetically)
    :rtype: tuple

    ..note: entries are sorted to insure consistency
    """
    def _get_entries(node, depth_):
        if isinstance(node, h5py.Dataset):
            return {}
        res_buf = {}
        if is_process_node(node) is True:
            res_buf[node.name] = int(node['sequence_index'][()])
        assert isinstance(node, h5py.Group)
        if depth_ >= 1:
            for sub_node in node.values():
                res_buf[node.name] = _get_entries(node=sub_node, depth_=depth_-1)
        return res_buf

    res = {}
    for node in root_node.values():
        res.update(_get_entries(node=node, depth_=depth-1))

    return res


def is_process_node(node):
    return (node.name.split('/')[-1].startswith('tomwer_process_') and
            'NX_class' in node.attrs and
            node.attrs['NX_class'] == "NXprocess" and
            'program' in node and
            h5py_read_dataset(node['program']) == 'tomwer_dark_refs' and
            'version' in node and
            'sequence_index' in node)


def get_darks_frm_process_file(process_file, entry) -> typing.Union[None, dict]:
    """

    :param process_file:
    :return:
    """
    if entry is None:
        with HDF5File(process_file, 'r', swmr=True) as h5f:
            entries = get_process_entries(root_node=h5f, depth=MAX_DEPTH)
            if len(entries) == 0:
                logger.info(
                    'unable to find a DarkRef process in %s' % process_file)
                return None
            elif len(entries) > 0:
                raise ValueError('several entry found, entry should be '
                                 'specify')
            else:
                entry = list(entries.keys())[0]
                logger.info('take %s as default entry' % entry)

    with HDF5File(process_file, 'r', swmr=True) as h5f:
        dark_nodes = get_process_entries(root_node=h5f[entry],
                                          depth=MAX_DEPTH-1)
        index_to_path = {}
        for key, value in dark_nodes.items():
            index_to_path[key] = key

        if len(dark_nodes) == 0:
            return {}
        # take the last processed dark ref
        last_process_index = sorted(dark_nodes.keys())[-1]
        last_process_dark = index_to_path[last_process_index]
        if(len(index_to_path)) > 1:
            logger.warning('several processing found for dark-ref,'
                           'take the last one: %s' % last_process_dark)

        res = {}
        if 'results' in h5f[last_process_dark].keys():
            results_node = h5f[last_process_dark]['results']
            if 'darks' in results_node.keys():
                darks = results_node['darks']
                for index in darks:
                    res[int(index)] = DataUrl(
                        file_path=process_file,
                        data_path=darks[index].name,
                        scheme="silx"
                    )
        return res


def get_flats_frm_process_file(process_file, entry) -> typing.Union[None, dict]:
    """

    :param process_file:
    :return:
    """
    if entry is None:
        with HDF5File(process_file, 'r', swmr=True) as h5f:
            entries = get_process_entries(root_node=h5f, depth=MAX_DEPTH)
            if len(entries) == 0:
                logger.info(
                    'unable to find a DarkRef process in %s' % process_file)
                return None
            elif len(entries) > 0:
                raise ValueError('several entry found, entry should be '
                                 'specify')
            else:
                entry = list(entries.keys())[0]
                logger.info('take %s as default entry' % entry)

    with HDF5File(process_file, 'r', swmr=True) as h5f:
        dkref_nodes = get_process_entries(root_node=h5f[entry],
                                          depth=MAX_DEPTH-1)
        if len(dkref_nodes) == 0:
            return {}
        index_to_path = {}
        for key, value in dkref_nodes.items():
            index_to_path[key] = key

        # take the last processed dark ref
        last_process_index = sorted(dkref_nodes.keys())[-1]
        last_process_dkrf = index_to_path[last_process_index]
        if(len(index_to_path)) > 1:
            logger.warning('several processing found for dark-ref,'
                           'take the last one: %s' % last_process_dkrf)

        res = {}
        if 'results' in h5f[last_process_dkrf].keys():
            results_node = h5f[last_process_dkrf]['results']
            if 'flats' in results_node.keys():
                flats = results_node['flats']
                for index in flats:
                    res[int(index)] = DataUrl(
                        file_path=process_file,
                        data_path=flats[index].name,
                        scheme="silx"
                    )
        return res



from os import path, mkdir
from ..resources.logger import LoggerOrPrint
from ..resources.utils import is_hdf5_extension
from ..utils import check_supported
from ..io.writer import Writers, NXProcessWriter
from ..io.utils import check_h5py_version # won't be necessary once h5py >= 3.0 required
from .params import files_formats
#
# Decorators and callback mechanism
#

def use_options(step_name, step_attr):
    def decorator(func):
        def wrapper(*args, **kwargs):
            self = args[0]
            if step_name not in self.processing_steps:
                self.__setattr__(step_attr, None)
                return
            self._steps_name2component[step_name] = step_attr
            self._steps_component2name[step_attr] = step_name
            return func(*args, **kwargs)
        return wrapper
    return decorator


def pipeline_step(step_attr, step_desc):
    def decorator(func):
        def wrapper(*args, **kwargs):
            self = args[0]
            if self.__getattribute__(step_attr) is None:
                return
            self.logger.info(step_desc)
            res = func(*args, **kwargs)
            self.logger.debug("End " + step_desc)
            step_name = self._steps_component2name[step_attr]
            callbacks = self._callbacks.get(step_name, None)
            if callbacks is not None:
                for callback in callbacks:
                    callback(self)
            if step_name in self._data_dump:
                self._dump_data_to_file(step_name)
            return res
        return wrapper
    return decorator


#
# Writer
#

class WriterConfigurator:

    _overwrite_warned = False

    def __init__(
        self, output_dir,
        file_prefix,
        file_format="hdf5",
        overwrite=False,
        start_index=None,
        logger=None,
        nx_info=None,
        write_histogram=False,
        histogram_entry="entry",
        writer_options=None,
        extra_options=None
    ):
        """
        Create a Writer from a set of parameters.

        Parameters
        ----------
        output_dir: str
            Directory where the file(s) will be written.
        file_prefix: str
            File prefix (without leading path)
        start_index: int, optional
            Index to start the files numbering (filename_0123.ext).
            Default is 0.
            Ignored for HDF5 extension.
        logger: nabu.resources.logger.Logger, optional
            Logger object
        nx_info: dict, optional
            Dictionary containing the nexus information.
        write_histogram: bool, optional
            Whether to also write a histogram of data. If set to True, it will configure
            an additional "writer".
        histogram_entry: str, optional
            Name of the HDF5 entry for the output histogram file, if write_histogram is True.
            Ignored if the output format is already HDF5 : in this case, nx_info["entry"] is taken.
        writer_options: dict, optional
            Other advanced options to pass to Writer class.
        """
        self.logger = LoggerOrPrint(logger)
        self.start_index = start_index
        self.write_histogram = write_histogram
        self.overwrite = overwrite
        writer_options = writer_options or {}
        self.extra_options = extra_options or {}

        check_supported(file_format, list(Writers.keys()), "output file format")

        self._set_output_dir(output_dir)
        self._set_file_name(file_prefix, file_format)

        # Init Writer
        writer_cls = Writers[file_format]
        writer_args = [self.fname]
        writer_kwargs = {}
        self._writer_exec_args = []
        self._writer_exec_kwargs = {}
        self._is_hdf5_output = is_hdf5_extension(file_format)

        if self._is_hdf5_output:
            writer_kwargs["entry"] = nx_info["entry"]
            writer_kwargs["filemode"] = "a"
            writer_kwargs["overwrite"] = overwrite
            self._writer_exec_args.append(nx_info["process_name"])
            self._writer_exec_kwargs["processing_index"] = nx_info["processing_index"]
            self._writer_exec_kwargs["config"] = nx_info["config"]
            check_h5py_version(self.logger)
        else:
            writer_kwargs["start_index"] = self.start_index
            if writer_options.get("tiff_single_file", False) and "tif" in file_format:
                do_append = writer_options.get("single_tiff_initialized", False)
                writer_kwargs.update({"multiframe": True, "append": do_append})

        if files_formats.get(file_format, None) == "jp2":
            cratios = self.extra_options.get("jpeg2000_compression_ratio", None)
            if cratios is not None:
                cratios = [cratios]
            writer_kwargs["cratios"] = cratios
            writer_kwargs["float_clip_values"] = self.extra_options.get("float_clip_values", None)
        self.writer = writer_cls(*writer_args, **writer_kwargs)

        if self.write_histogram and not(self._is_hdf5_output):
            self._init_separate_histogram_writer(histogram_entry)


    def _set_output_dir(self, output_dir):
        self.output_dir = output_dir
        if path.exists(self.output_dir):
            if not(path.isdir(self.output_dir)):
                raise ValueError(
                    "Unable to create directory %s: already exists and is not a directory"
                    % self.output_dir
                )
        else:
            self.logger.debug("Creating directory %s" % self.output_dir)
            mkdir(self.output_dir)


    def _set_file_name(self, file_prefix, file_format):
        self.file_prefix = file_prefix
        self.file_format = file_format
        self.fname = path.join(
            self.output_dir,
            file_prefix + "." + file_format
        )
        if path.exists(self.fname):
            err = "File already exists: %s" % self.fname
            if self.overwrite:
                if not(WriterConfigurator._overwrite_warned):
                    self.logger.warning(err + ". It will be overwritten as requested in configuration")
                    WriterConfigurator._overwrite_warned = True
            else:
                self.logger.fatal(err)
                raise ValueError(err)


    def _init_separate_histogram_writer(self, hist_entry):
        hist_fname = path.join(
            self.output_dir,
            "histogram_%04d.hdf5" % self.start_index
        )
        self.histogram_writer = NXProcessWriter(
            hist_fname,
            entry=hist_entry,
            filemode="w",
            overwrite=True,
        )


    def get_histogram_writer(self):
        if not(self.write_histogram):
            return None
        if self._is_hdf5_output:
            return self.writer
        else:
            return self.histogram_writer


    def write_data(self, data):
        self.writer.write(
            data,
            *self._writer_exec_args,
            **self._writer_exec_kwargs
        )



import numpy as np
import pytest
from scipy.misc import ascent
from nabu.misc.unsharp import UnsharpMask
from nabu.misc.unsharp_opencl import OpenclUnsharpMask, __have_opencl__
from nabu.cuda.utils import __has_pycuda__
if __have_opencl__:
    import pyopencl.array as parray
if __has_pycuda__:
    import pycuda.gpuarray as garray
    from nabu.misc.unsharp_cuda import CudaUnsharpMask

@pytest.fixture(scope='class')
def bootstrap(request):
    cls = request.cls
    cls.data = np.ascontiguousarray(ascent()[:, :511], dtype=np.float32)
    cls.tol = 1e-4
    cls.sigma = 1.6
    cls.coeff = 0.5
    cls.compute_reference()


@pytest.mark.usefixtures('bootstrap')
class TestUnsharp:

    @classmethod
    def compute_reference(cls):
        cls.Unsharp = UnsharpMask(cls.data.shape, cls.sigma, cls.coeff)
        cls.ref = cls.Unsharp.unsharp(cls.data)

    @pytest.mark.skipif(not(__have_opencl__), reason="Need pyopencl for this test")
    def testOpenclUnsharp(self):
        ClUnsharp = OpenclUnsharpMask(self.data.shape, self.sigma, self.coeff)
        d_image = parray.to_device(ClUnsharp.queue, self.data)
        d_out = parray.zeros_like(d_image)
        ClUnsharp.unsharp(d_image, d_out)
        mae = np.max(np.abs(d_out.get() - self.ref))
        assert mae < self.tol, "Max error is too high (%.2e > %.2e)" % (mae, self.tol)

    @pytest.mark.skipif(not(__has_pycuda__), reason="Need cuda/pycuda for this test")
    def testCudaUnsharp(self):
        CuUnsharp = CudaUnsharpMask(self.data.shape, self.sigma, self.coeff)
        d_image = garray.to_gpu(self.data)
        d_out = garray.zeros_like(d_image)
        CuUnsharp.unsharp(d_image, d_out)
        mae = np.max(np.abs(d_out.get() - self.ref))
        assert mae < self.tol, "Max error is too high (%.2e > %.2e)" % (mae, self.tol)



import pytest
import numpy as np
from scipy.ndimage import shift as ndshift
from nabu.preproc.shift import VerticalShift
from nabu.cuda.utils import __has_pycuda__, get_cuda_context
if __has_pycuda__:
    from nabu.preproc.shift_cuda import CudaVerticalShift, garray

@pytest.fixture(scope='class')
def bootstrap(request):
    cls = request.cls
    data = np.zeros([13, 11], "f")
    slope = 100 + np.arange(13)
    data[:] = slope[:, None]
    cls.radios = np.array([data] * 17)
    cls.shifts = 0.3 + np.arange(17)
    cls.indexes = range(17)
    # given the shifts and the radios we build the golden reference
    golden = []
    for iradio in range(17):
        projection_number = cls.indexes[iradio]
        my_shift = cls.shifts[projection_number]
        padded_radio = np.concatenate(
            [cls.radios[iradio], np.zeros([1, 11], "f")], axis=0
        )  # needs padding because ndshifs does not work as expected
        shifted_padded_radio = ndshift(
            padded_radio, [-my_shift, 0], mode="constant", cval=0.0, order=1
        ).astype("f")
        shifted_radio = shifted_padded_radio[:-1]
        golden.append(shifted_radio)
    cls.golden = np.array(golden)
    cls.tol = 1e-5
    if __has_pycuda__:
        cls.ctx = get_cuda_context()

@pytest.mark.usefixtures('bootstrap')
class TestVerticalShift:
    def test_vshift(self):
        radios = self.radios.copy()
        new_radios = np.zeros_like(radios)

        Shifter = VerticalShift(radios.shape, self.shifts)

        Shifter.apply_vertical_shifts(radios, self.indexes, output=new_radios)
        assert abs(new_radios - self.golden).max() < self.tol


        Shifter.apply_vertical_shifts(radios, self.indexes)
        assert abs(radios - self.golden).max() < self.tol


    @pytest.mark.skipif(not(__has_pycuda__), reason="Need cuda/pycuda for this test")
    def test_cuda_vshift(self):
        d_radios = garray.to_gpu(self.radios)
        d_out = garray.zeros_like(d_radios)

        Shifter = CudaVerticalShift(d_radios.shape, self.shifts)

        Shifter.apply_vertical_shifts(d_radios, self.indexes, output=d_out)
        assert abs(d_out.get() - self.golden).max() < self.tol

        Shifter.apply_vertical_shifts(d_radios, self.indexes)
        assert abs(d_radios.get() - self.golden).max() < self.tol



from os import getpid
from datetime import datetime
from socket import gethostname


# Default formatter
FORMATTER = "[{hostname}/{pid}/{worker}] {day}/{month}/{year} {hour}:{min}:{sec} {message}"


def format_message(message, worker_name=None, replacements=None):
    """
    Format a message with additional information.
    The formating is defined by `nabu.distributed.utils.FORMATTER`.

    Parameters
    -----------
    message: str
        Message to log.
    worker_name: str, optional
        Name of the current worker.
    replacements: dict, optional
        Dictionary of additional patterns that should be replaced in the current
        formatter.
    """
    now = datetime.now()
    # poor man's templating engine
    formatters = {
        "hostname": gethostname(),
        "pid": getpid(),
        "worker": "",
        "day": now.strftime("%d"),
        "month": now.strftime("%m"),
        "year": now.strftime("%Y"),
        "hour": now.strftime("%H"),
        "min": now.strftime("%M"),
        "sec": now.strftime("%S"),
        "message": message
    }
    if worker_name is not None:
        formatters["worker"] = worker_name
    FMT = FORMATTER.replace("{", "").replace("}", "")
    for fmt_key, fmt_val in formatters.items():
        FMT = FMT.replace(fmt_key, str(fmt_val))
    if replacements is not None:
        for what, replacement in replacements.items():
            FMT = FMT.replace(what, replacement)
    return FMT


def log(message):
    """
    Log a message using the current formater.
    """
    print(format_message(message))






from ..pipeline.fullfield.local_reconstruction import ChunkedReconstructor, GroupedReconstructor
from ..utils import deprecation_warning

# Compat.
LocalReconstruction = ChunkedReconstructor
FullFieldReconstructor = ChunkedReconstructor
FullRadiosReconstructor = GroupedReconstructor
#

deprecation_warning(
    "nabu.resources.app.local_reconstruction was moved to nabu.pipeline.fullfield.local_reconstruction.",
    func_name="fullfield_local_reconstruction"
)